

__author__ = "Andrea Tramacere"


import matplotlib as mpl


try:
    from matplotlib import  pyplot as plt
except:
    try:
        from matplotlib import pylab as plt

    except:
        try:
           import  pylab as plt
        except:
            raise RuntimeError('Unable to import pylab/pyplot from matplotlib')



from matplotlib import gridspec
import numpy as np
import  os
from astropy.constants import m_e,m_p,c
import matplotlib.ticker as ticker
import warnings

from collections import namedtuple

from .output import section_separator,WorkPlace

from .utils import *

__all__=['PlotSED','BasePlot','PlotPdistr','PlotSpecComp','PlotSeedPhotons','PlotSpectralMultipl','PlotTempEvDiagram','PlotTempEvEmitters']

def y_ev_transf(x):
    return x / 2.417E14

def y_ev_transf_inv(x):
    return x * 2.417E14



def set_mpl():
    mpl.rcParams['figure.figsize'] = [12.0, 8.0]
    mpl.rcParams['figure.dpi'] = 100
    mpl.rcParams['savefig.dpi'] = 100

    mpl.rcParams['font.size'] = '14'
    mpl.rcParams['legend.fontsize'] = 'medium'
    mpl.rcParams['figure.titlesize'] = 'medium'



def _rescale( x_min=None, x_max=None, y_min=None, y_max=None):
        warnings.warn('`The rescale method has been removed and has been replaced by the setlim method')
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        print("!The rescale method as been replaced by the setlim method            !")
        print("!please notice that now jetset uses log axis rather than loglog plots!")
        print("!so, the correct way to use it is rescale(x_min=8)->setlim(x_min=1E8)!")
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")

class  PlotSED (object):
    def __init__(self,
                 sed_data=None,
                 model=None,
                 interactive=False,
                 plot_workplace=None,
                 title='Plot',
                 frame='obs',
                 density=False,
                 dpi=100,
                 figsize=(12,8),
                 use_grid=True):

        check_frame(frame)

        self.frame=frame

        self.axis_kw=['x_min','x_max','y_min','y_max']
        self.interactive=interactive

        plot_workplace=plot_workplace
        self.lines_data_list=[]
        self.lines_model_list=[]
        self.lines_res_list = []

        if self.interactive is True:
            plt.ion()
            print ('running PyLab in interactive mode')

        if plot_workplace is None:
            plot_workplace=WorkPlace()
            self.out_dir=plot_workplace.out_dir
            self.flag=plot_workplace.flag

        else:
            self.out_dir=plot_workplace.out_dir
            self.flag=plot_workplace.flag


            self.title="%s_%s"%(title,self.flag)

        if figsize is None:
            figsize=(10,8)

        self.fig=plt.figure(figsize=figsize,dpi=dpi)

        self.gs = gridspec.GridSpec(2, 1, height_ratios=[4, 1])

        self.sedplot= self.fig.add_subplot(self.gs[0])
        self._add_res_plot()

        self.set_plot_axis_labels(density=density)

        #if autoscale==True:
        self.sedplot.set_autoscalex_on(True)
        self.sedplot.set_autoscaley_on(True)
        self.sedplot.set_autoscale_on(True)
        self.counter=0
        if use_grid is True:
            self.sedplot.grid(use_grid,alpha=0.5)

        self.sedplot.set_xlim(1E6, 1E30)
        if frame == 'obs':
            self.sedplot.set_ylim(1E-20, 1E-8)
        elif frame == 'src':
            self.sedplot.set_ylim(1E38, 1E55)
        elif frame == 'blob':
            self.sedplot.set_ylim(1E34, 1E51)
        else:
            unexpected_behaviour()

        self.sedplot.set_xscale("log", nonpositive='clip')
        self.sedplot.set_yscale("log", nonpositive='clip')
   
        self.secaxy = self.sedplot.secondary_xaxis('top', functions=(y_ev_transf, y_ev_transf_inv))
        self.secaxy.set_xlabel('E (eV)')

        self.resplot.set_ybound(-2,2)
        
        try:
            if hasattr(self.fig.canvas.manager,'toolbar'):
                self.fig.canvas.manager.toolbar.update()
        except:
            pass

        if sed_data is not None :
            self.add_data_plot(sed_data,density=density)

        if model is not  None:
            self.add_model_plot(model)
        self.counter_res=0

        self.add_residual_plot = self.add_model_residual_plot

    def _check_frame(self,frame):
        if frame is None:
            frame=self.frame

        elif frame != self.frame:
            raise RuntimeError('you have to use the same restframe of the PlotSED class:',self.frame )

        return frame

    def _add_res_plot(self):
        self.resplot = self.fig.add_subplot(self.gs[1], sharex=self.sedplot)

        self.lx_res = '$ \\nu $  (Hz)'
        self.ly_res = 'res'

        self.resplot.set_ylabel(self.ly_res)
        self.resplot.set_xlabel(self.lx_res)
        #self,resplot.set_xscale("log", nonpositive='clip')
        self.add_res_zeroline()

    def clean_residuals_lines(self):
        for i in range(len(self.lines_res_list)):
            self.del_residuals_line(0)

    def clean_data_lines(self):

        for i in range(len(self.lines_data_list)):
            self.del_data_line(0)

    def clean_model_lines(self):
        for i in range(len(self.lines_model_list)):
            self.del_model_line(0)


    def list_lines(self):
        if self.lines_data_list==[] and self.lines_model_list==[]:
            pass
        else:

            for ID,plot_line in enumerate(self.lines_data_list):
                print('data',ID, plot_line.get_label())

            for ID,plot_line in enumerate(self.lines_model_list):
                print ('model',ID,  plot_line.get_label())

    def del_data_line(self,line_ID):
        if self.lines_data_list==[]:
            print  ("no lines to delete ")
        else:
            print ("removing line: ",self.lines_data_list[line_ID])
            line = self.lines_data_list[line_ID]

            for item in line:
                # This removes lines
                if np.shape(item) == ():
                    item.remove()
                else:
                    # This removes containers for data with errorbars
                    for item1 in item:
                        item1.remove()

            del self.lines_data_list[line_ID]
            #self.update_legend()
            self.update_plot()

    def del_model_line(self,line_ID):

        if self.lines_model_list==[]:
            #print  "no lines to delete "
            pass
        else:

            line=self.lines_model_list[line_ID]
            line.remove()


            del self.lines_model_list[line_ID]

            self.update_plot()
            #self.update_legend()

    def del_residuals_line(self, line_ID):
        if self.lines_res_list == []:
            # print  "no lines to delete "
            pass
        else:

            line = self.lines_res_list[line_ID]
            line.remove()

            del self.lines_res_list[line_ID]

            self.update_plot()
            #self.update_legend()

    def set_plot_axis_labels(self, density=False):
        self.lx = '$ \\nu $  (Hz)'

        if self.frame == 'src' or self.frame == 'blob':

            if density is False:
                self.ly = '$ \\nu L_{\\nu} $   (erg  s$^{-1})$'
            else:
                self.ly = '$   L_{\\nu} $   (erg  s$^{-1}$ Hz$^{-1})$'

        elif self.frame == 'obs':
            if density is False:
                self.ly = '$ \\nu F_{\\nu} $   (erg cm$^{-2}$  s$^{-1})$'
            else:
                    self.ly = '$   F{\\nu} $   (erg cm$^{-2}$  s$^{-1}$ Hz$^{-1})$'


        else:
            unexpected_behaviour()

        self.sedplot.set_ylabel(self.ly)
        self.sedplot.set_xlabel(self.lx)



    def add_res_zeroline(self):
        #y0 = np.zeros(2)
        #x0 = [0,30]
        self.resplot.axhline(0, ls='--', color='black')
        self.update_plot()

    
    def rescale(self, x_min=None, x_max=None, y_min=None, y_max=None):
        _rescale(x_min=x_min,x_max=x_max,y_min=y_min,y_max=y_max)

    def setlim(self, x_min=None, x_max=None, y_min=None, y_max=None):
        self.sedplot.set_xlim(x_min, x_max)
        self.sedplot.set_ylim(y_min, y_max)
    
    def setlim_res(self,x_min=None,x_max=None,y_min=None,y_max=None):
        self.resplot.set_xlim(x_min,x_max)
        self.resplot.set_ylim(y_min,y_max)
        self.update_plot()

    def update_plot(self):
        self.fig.canvas.draw()
       
        y_s = []
        x_min = []
        x_max = []
        y_min = None
        y_max = None
        if len(self.sedplot.lines)>0:

            for l in self.sedplot.lines:
                if len(l.get_ydata())>0:
                    y_s.append(np.max(l.get_ydata()))
            if len(y_s) > 0:
                y_min = min(y_s)/1000
                y_max = max(y_s)*10
            else:
                self.sedplot.autoscale(axis='y')
            if y_min is not None and y_max is not None:
                self.sedplot.set_ylim(y_min, y_max)
                for l in self.sedplot.lines:
                    x=np.array(l.get_xdata())[np.array(l.get_ydata()) >= y_min]
                    if len(x)>0:
                        x_min.append(np.min(x))
                        x_max.append(np.max(x))
                if len(x_min)>0  and  len(x_max)>0:
                    self.sedplot.set_xlim(min(x_min)/10, max(x_max)*10)
        else:
            self.sedplot.relim()
            self.sedplot.autoscale(axis='y')
            self.sedplot.autoscale(axis='x')
    
        self.update_legend()
        self.fig.tight_layout()

    def update_legend(self,label=None):

        _handles=[]

        if self.lines_data_list!=[] and self.lines_data_list is not None:
            _handles.extend(self.lines_data_list)

        if self.lines_model_list!=[] and self.lines_model_list is not None:
            _handles.extend(self.lines_model_list)

        for h in _handles[:]:
            if h._label is  None:
                _handles.remove(h)
            elif h._label.startswith('_line'):
                _handles.remove(h)
            else:
                 pass

        self.sedplot.legend(handles=_handles,loc='center left', bbox_to_anchor=(1.0, 0.5), ncol=1, prop={'size':10})




    def add_model_plot(self, model, label=None, color=None, line_style=None, flim=None,auto_label=True,fit_range=None,density=False, update=True, lw=1.0 ,frame=None):

        frame=self._check_frame(frame=frame)

        if hasattr(model,'get_model_points'):
            try:
                x, y = model.get_model_points(log_log=False, frame = self.frame)
            except Exception as e:
                raise RuntimeError('for model',model.name, "problem with get_model_points()",e)
        else:
            try:
                x, y = model.SED.get_model_points(log_log=False, frame = self.frame)
            except Exception as e:
                raise RuntimeError('for model',model.name, "problem with SED.get_model_points()",e)

        if density is True:
            y=y/x
        if line_style is None:
            line_style = '-'

        if label is None and auto_label is True:
            if model.name is not None:
                label = model.name
            else:
                label = 'line %d' % self.counter

        if flim is not None:
            msk=y>flim
            x=x[msk]
            y=y[msk]
        else:
            pass

        if fit_range is not None:
            msk1 = x < fit_range[1]
            msk2 = x > fit_range[0]

            x = x[msk1 * msk2]
            y = y[msk1 * msk2]

        line, = self.sedplot.plot(x, y, line_style, label=label,color=color,lw=lw)


        self.lines_model_list.append(line)

        if update is True:
            #self.update_legend()
            self.update_plot()

        self.counter += 1

    def plot_tempev_model(self,
                          temp_ev,
                          region,
                          comp='Sum',
                          frame=None,
                          t1=None,
                          t2=None,
                          time_slice=None,
                          time_slice_bin=None,
                          time=None,
                          time_bin=None,
                          density=False,
                          use_cached=False,
                          sed_data=None,
                          average=False):

        frame=self._check_frame(frame)


        if (time_slice is not None and time is not None):
            raise RuntimeError('you can to pass either the N-th time slice "time_slice", or the blob time in seconds "time" ')

        if t1 is None or t1 < region.time_sampled_emitters.time_blob[0]:
            t1 = region.time_sampled_emitters.time_blob[0]

        if t2 is None or t2 > region.time_sampled_emitters.time_blob[-1]:
            t2 = region.time_sampled_emitters.time_blob[-1]

        if time_slice is None:
            _time_slice = 0
        else:
            _time_slice = time_slice

        _time_slice_bin = time_slice_bin
        if time_slice_bin is None and time_slice is None:
            _time_slice_bin = 1


        #if time_slice is not None or time_bin is None:

        if time is not None and time_bin is not None:
            t_array = np.arange(t1, t2, time_bin)
            time_id_array=None
        elif time is not None and time_bin is None:
            t_array = np.array([time])
            time_id_array = None
        else:
            t_array, time_id_array = region.time_sampled_emitters._get_time_samples(time_slice=_time_slice,
                                                                                    time_slice_bin=_time_slice_bin)
            time_id_array = time_id_array[t_array <= t2]
            time_id_array = time_id_array[t_array >= t1]
            t_array = t_array[t_array <= t2]
            t_array = t_array[t_array >= t1]


        g = plt.cm.Greens(np.linspace(0.5, 1, t_array.size))
        r = plt.cm.Reds(np.linspace(0.5, 1, t_array.size))
        b = plt.cm.Blues(np.linspace(0.5, 1, t_array.size))
        for ID, t in enumerate(t_array):
            if time is not None:
                s = region.get_SED(comp, frame=frame, time=t, use_cached=use_cached, time_bin=time_bin,average=average)
            else:
                s = region.get_SED(comp, frame=frame, time_slice=time_id_array[ID], use_cached=use_cached,
                                   time_slice_bin=time_slice_bin,average=average)

            label = None
            ls = '-'
            color = r[ID]
            if temp_ev.custom_q_jnj_profile[temp_ev._get_time_slice_T_array(t)] > 0:
                color = g[ID]
                ls = '-'
                lw = 0.2
            if temp_ev.custom_acc_profile[temp_ev._get_time_slice_T_array(t)] > 0:
                color = b[ID]
                ls = '-'
                lw = 0.2

            if ID == 0:
                lw = 2
                ls = '--'
                label = 'start, t=%2.2e (s)' % t
                color = 'green'
            if ID == t_array.size - 1:
                lw = 2
                ls = '--'
                color = 'purple'
                label = 'stop, t=%2.2e (s)' % t

            self.add_model_plot(model=s, label=label, line_style=ls, color=color, update=False, lw=lw,
                                    auto_label=False,density=density)

        if sed_data is not None:
            self.add_data_plot(sed_data)

        self.update_plot()
        return


    def add_data_plot(self,sed_data,label=None,color=None,frame=None,fmt='o',ms=4,mew=0.5,fit_range=None, density = False):

        frame = self._check_frame(frame)
        try:
            x,y,dx,dy,=sed_data.get_data_points(log_log=False,frame=self.frame, density=density)
        except Exception as e:
            raise RuntimeError("!!! ERROR failed to get data points from", sed_data,e)


        if dx is None:
            dx=np.zeros(len(sed_data.data['nu_data']))


        if dy is None:
            dy=np.zeros(len(sed_data.data['nu_data']))

        UL = sed_data.data['UL']

        if label is None:
            if sed_data.obj_name is not None  :
                label=sed_data.obj_name
            else:
                label='line %d'%self.counter

        if fit_range is not None:
            msk1 = x < fit_range[1]
            msk2 = x > fit_range[0]

            x = x[msk1 * msk2]
            y = y[msk1 * msk2]
            dx= dx[msk1 * msk2]
            dy = dy[msk1 * msk2]
            UL=UL[msk1 * msk2]

        line = self.sedplot.errorbar(x, y, xerr=dx, yerr=dy, fmt=fmt
                                     , uplims=UL,label=label,ms=ms,mew=mew,color=color)

        self.lines_data_list.append(line)

        self.counter+=1
        #self.update_legend()
        self.update_plot()
    


    def add_xy_plot(self,x,y,label=None,color=None,line_style=None,autoscale=False):

        if line_style is None:
            line_style='-'


        if label is None:
            label='line %d'%self.counter

        line, = self.sedplot.plot(x, y, line_style,label=label)

        self.lines_model_list.append(line)

        self.counter+=1

        #self.update_legend()
        self.update_plot()



    def add_model_residual_plot(self, model, data, label=None, color=None, filter_UL=True, fit_range=None):
        if data is not None:
            x,y = model.get_residuals(log_log=False,data=data,filter_UL=filter_UL)
            self.add_xy_residual_plot(x=x, y=y, fit_range=fit_range, color=color)
        else:
            pass


    def add_xy_residual_plot(self, x, y, fit_range=None, color=None):
        if self.counter_res == 0:
            self.add_res_zeroline()
        if fit_range is not None:
            msk1 = x < fit_range[1]
            msk2 = x > fit_range[0]

            x = x[msk1 * msk2]
            y = y[msk1 * msk2]

        line = self.resplot.errorbar(x, y, yerr=np.ones(x.size), fmt='+', color=color)
        self.lines_res_list.append(line)
        self.counter_res += 1
        self.update_plot()


    def add_text(self,lines):
        self.PLT.focus(0,0)
        x_min, x_max = self.sedplot.get_xlim()
        y_min, y_max = self.sedplot.get_ylim()
        t=''
        for line in lines:
            t+='%s \\n'%line.strip()
        self.PLT.text(t,font=10,charsize=0.6,x=x_min-1.5,y=y_min-2.85)
        self.PLT.redraw()


    def save(self,filename=None):
        if filename is None:
            wd=self.out_dir
            filename = 'jetset_fig.png'

        else:
            wd=''

        outname = os.path.join(wd,filename)
        self.fig.savefig(outname)

    def show(self):
        self.fig.show()





class BasePlot(object):

    def __init__(self,figsize=(8,6),dpi=100):
        self.fig, self.ax = plt.subplots(figsize=figsize,dpi=dpi)

    def rescale(self, x_min=None, x_max=None, y_min=None, y_max=None):
        _rescale(x_min=x_min,x_max=x_max,y_min=y_min,y_max=y_max)

    def setlim(self, x_min=None, x_max=None, y_min=None, y_max=None):
        self.ax.set_xlim(x_min, x_max)
        self.ax.set_ylim(y_min, y_max)

    def update_plot(self):
        self.fig.canvas.draw()
        self.ax.relim()
        self.ax.autoscale(axis='y')
        self.ax.legend()
        self.fig.tight_layout()



class PlotSpectralMultipl(BasePlot):
    def __init__(self):
        super(PlotSpectralMultipl, self).__init__()

        secax = self.ax.secondary_xaxis('top', functions=(y_ev_transf, y_ev_transf_inv))
        secax.set_xlabel('E (eV)')


    def plot(self,nu,y,y_label,y_min=None,y_max=None,label=None,line_style=None,color=None):

        self.ax.plot(np.log10(nu), np.log10(y),label=label,ls=line_style,color=color)
        self.ax.set_xlabel(r'$ \nu $  (Hz)')
        self.ax.set_ylabel(y_label)
        self.ax.set_ylim(y_min, y_max)
        self.ax.legend()
        self.update_plot()




class  PlotPdistr (BasePlot):

    def __init__(self,figsize=(8,6),dpi=100,injection=False,loglog=True):
        super(PlotPdistr, self).__init__(figsize=figsize,dpi=dpi)
        self.loglog=loglog
        self.injection = injection

    def _set_variable(self,gamma,n_gamma,particle,energy_unit,pow=None):

        energy_plot=False
        if energy_unit == 'gamma':
            energy_name = '\gamma'
            energy_units=''
        else:
            energy_name='E'
            energy_units= '%s'%energy_unit
            energy_plot=True

        if  energy_plot is False:
            x=gamma
            y=n_gamma

        else:

            if particle=='electrons':
                x = gamma*(m_e*c*c).to(energy_unit).value
                y = n_gamma * 1.0/(m_e*c*c).to(energy_unit).value
            elif particle=='protons':
                x = gamma * (m_p * c * c).to(energy_unit).value
                y = n_gamma * 1.0 / (m_p * c * c).to(energy_unit).value
            else:
                raise  RuntimeError('particle ',particle, 'not implemented')

        m = y > 0
        x=np.copy(x)
        y=np.copy(y)

        if pow is not None:
            y[m] = y[m]* np.power( x[m], pow)

        if self.loglog is True:
            x[m] = np.log10( x[m])
            y[m] = np.log10(y[m])

        return x[m], y[m], energy_name,energy_units



    def _set_xy_label(self,energy_name,energy_units,pow):
        if energy_units != '':
            _e = '(%s)' % energy_units
        else:
            _e = ''

        if self.loglog is True:
            self.ax.set_xlabel(r'log($%s$)  %s' % (energy_name, _e))
        else:
            self.ax.set_xlabel(r'$%s$  %s' % (energy_name, _e))

        if energy_units != '':
            _e = '%s^{-1}' % energy_units
        else:
            _e = ''

        n_str = 'n($%s$)'%energy_name
        if pow is not None and pow!=0:
            n_str = 'n($%s$) $%s^{%d}$' % (energy_name,energy_name,pow)
            if energy_units != '':
                i=-1+pow
                if i==1:
                    _e='%s' %(energy_units)
                else:
                    _e='%s^{%d}' %(energy_units,i)
        if self.injection is False:

            if self.loglog is True:
                self.ax.set_ylabel(r'log(%s)   ($cm^{-3} %s$) ' % (n_str, _e))
            else:
                self.ax.set_ylabel(r'%s   ($cm^{-3} %s$) ' % (n_str, _e))
        else:
            if self.loglog is True:
                self.ax.set_ylabel(r'log(Q$_{inj}$($%s$))   ($cm^{-3} s^{-1} %s$)' % (energy_name,_e))
            else:
                self.ax.set_ylabel(r'Q$_{inj}$($%s$)   ($cm^{-3} s^{-1} %s$)' % (energy_name,_e))

    def _plot(self,x,y,c=None,lw=None,ls=None,label=None):
        if self.loglog is True:
            self.ax.plot(x, y, c=c, lw=lw, label=label,ls=ls)
        else:
            self.ax.loglog(x, y,c=c, lw=lw, label=label,ls=ls)

    def plot_distr(self,gamma,n_gamma,y_min=None,y_max=None,x_min=None,x_max=None,particle='electrons',energy_unit='gamma',label=None):

        x,y,energy_name,energy_units=self._set_variable(gamma,n_gamma,particle,energy_unit)

        if label is None:
            label=particle
        self._plot(x,y,label=label)
        self._set_xy_label(energy_name,energy_units,pow=None)
        self.update_plot()
        self.ax.set_ylim(y_min, y_max)
        self.ax.set_xlim(x_min, x_max)


    def plot_distr2p(self, gamma, n_gamma, y_min=None, y_max=None, x_min=None, x_max=None,particle='electrons',energy_unit='gamma',label=None):
        if label is None:
            label=particle

        x, y, energy_name, energy_units = self._set_variable(gamma, n_gamma, particle, energy_unit,pow=2)
        self._plot(x,y,label=label)
        self._set_xy_label(energy_name, energy_units,pow=2)
        self.update_plot()
        self.ax.set_ylim(y_min, y_max)
        self.ax.set_xlim(x_min, x_max)

    def plot_distr3p(self,gamma,n_gamma,y_min=None,y_max=None,x_min=None,x_max=None,particle='electrons',energy_unit='gamma', label=None):
        if label is None:
            label = particle

        x, y, energy_name, energy_units = self._set_variable(gamma, n_gamma, particle, energy_unit, pow=3)
        self._plot(x,y,label=label)
        self._set_xy_label(energy_name, energy_units,pow=3)
        self.update_plot()
        self.ax.set_ylim(y_min, y_max)
        self.ax.set_xlim(x_min, x_max)


    def update_plot(self):
        self.fig.canvas.draw()
        self.ax.relim()
        self.ax.autoscale(axis='y')
        self.ax.autoscale(axis='x')
        self.ax.legend()
        self.fig.tight_layout()







class  PlotTempEvEmitters (PlotPdistr):

    def __init__(self,figsize=(8,6),dpi=100,loglog=True):
        super(PlotTempEvEmitters, self).__init__(figsize=figsize,dpi=dpi,loglog=loglog,)


    def _plot_distr(self,temp_ev,
                        region,particle='electrons',
                        energy_unit='gamma',
                        pow=None,
                        plot_Q_inj=True,
                        t1=None, 
                        t2=None,
                        #time_slice_bin=None,
                        #time_slice=None):
                    ):

        if t1 is None:
            t1=region.time_sampled_emitters.time_blob[0]

        if t2 is None:
            t2=region.time_sampled_emitters.time_blob[-1]
        
       # if time_slice_bin is None and time_slice is None:
       #     _time_slice_bin = 1

        t_array = np.linspace(t1, t2, region.time_sampled_emitters.time_blob.size,)

        ls = '-'
        lw = 0.2

        g = plt.cm.Greens(np.linspace(0.5, 1, t_array.size))
        r = plt.cm.Reds(np.linspace(0.5, 1, t_array.size))
        b = plt.cm.Blues(np.linspace(0.5, 1, t_array.size))

        n = region.time_sampled_emitters.n_gamma
        for ID, t in enumerate(t_array):
            label = None
            ls = '-'
            color = r[ID]

            if temp_ev.custom_q_jnj_profile[temp_ev._get_time_slice_T_array(t)] > 0:
                color = g[ID]
                ls = '-'
                lw = 0.2

            if temp_ev.custom_acc_profile[temp_ev._get_time_slice_T_array(t)] > 0:
                color = b[ID]
                ls = '-'
                lw = 0.2

            if ID == 0:
                lw = 2
                ls = '--'
                label = 'start, t=%2.2e (s)' % t
                color = 'green'
            if ID == t_array.size - 1:
                lw = 2
                ls = '--'
                color = 'purple'
                label = 'stop, t=%2.2e (s)' % t

            x, y, energy_name, energy_units = self._set_variable(region.time_sampled_emitters.gamma, n[ID], particle, energy_unit, pow=pow)
            self._plot(x,y,c=color,lw=lw,label=label,ls=ls)
        #x, y, energy_name, energy_units = self._set_variable(region.time_sampled_emitters.gamma, n[0], particle, #energy_unit, pow=pow)
        #self._plot(x, y, c='black', lw=2,label='Start sample')

        #x, y, energy_name, energy_units = self._set_variable(region.time_sampled_emitters.gamma, n[-1], particle, #energy_unit, pow=pow)
        #self._plot(x, y, c='blue', lw=2,label='Stop sample')
        self._set_xy_label(energy_name, energy_units,pow=pow)
        #TODO move to plot inj if iny is used in region
        if temp_ev.Q_inj is not None and (region._region_type == 'acc' or temp_ev._only_radiation is True):
            y = temp_ev.Q_inj.n_gamma_e * temp_ev.delta_t
            x = temp_ev.Q_inj.gamma_e
            if plot_Q_inj is True:
                if pow is not None:
                    y=y*np.power(x,pow)

                self._plot(x,y, c='red', lw=1, label='$Q_{inj}$ delta t')
        #print('==> d')

        self.ax.legend()

    def plot_distr(self, temp_ev, region='acc', energy_unit='gamma',plot_Q_inj=True,pow=None):
        self._plot_distr(temp_ev,region=region, particle='electrons',energy_unit=energy_unit,pow=pow,plot_Q_inj=plot_Q_inj)

    def plot_distr2p(self, temp_ev, region='acc', energy_unit='gamma',plot_Q_inj=True):
        self._plot_distr(temp_ev, region=region, particle='electrons',energy_unit=energy_unit,pow=2,plot_Q_inj=plot_Q_inj)

    def plot_distr3p(self, temp_ev, region='acc', energy_unit='gamma',plot_Q_inj=True):
        self._plot_distr(temp_ev, region=region, particle='electrons',energy_unit=energy_unit,pow=3,plot_Q_inj=plot_Q_inj)


class  PlotTempEvDiagram (object):

    def __init__(self,figsize=(8,6),dpi=100,expanding_region=False):
        if expanding_region is True:
            n_rows =4
        else:
            n_rows = 4
        self.fig, self.axs = plt.subplots(n_rows, 1, figsize=figsize, dpi=dpi, sharex=False)

    def plot(self,
             T_array,
             inj_profile,
             acc_profile,
             R_exp,
             B_exp,
             R_H_exp):


        self.axs[0].plot(T_array, acc_profile, label='Acc. start/stop', c='g')
        self.axs[0].set_ylim(0, 1.5)

        self.axs[1].plot(T_array, inj_profile, label='Inj. profile', c='b')
        self.axs[1].set_ylim(0,None)
        self.axs[1].set_xlabel('Time in blob frame (s)')
        self.axs[1].sharex = self.axs[0]
        #if expanding_region is True:
        self.axs[2].plot(np.log10(R_H_exp), np.log10(R_exp), label='Rad. region size', c='orange')
        self.axs[2].set_ylabel('log(R_rad) (cm)')
        self.axs[3].plot(np.log10(R_H_exp), np.log10(B_exp), label='B Rad. region', c='green')
        self.axs[3].set_xlabel('log(BH distance) in observer frame (cm)')
        self.axs[3].set_ylabel('log(B) (G)')
        self.axs[3].sharex = self.axs[2]

        for ax in self.axs:
            ax.legend()


        #self.fig.subplots_adjust(hspace=0)
        self.fig.tight_layout()

    def rescale(self, x_min=None, x_max=None, y_min=None, y_max=None):
        _rescale(x_min=x_min,x_max=x_max,y_min=y_min,y_max=y_max)

    def setlim(self, x_min=None, x_max=None, y_min=None, y_max=None):
        self.ax.set_xlim(x_min, x_max)
        self.ax.set_ylim(y_min, y_max)

class  PlotSpecComp (BasePlot):

    def __init__(self):
        super(PlotSpecComp, self).__init__()

    def plot(self,nu,nuFnu,y_min=None,y_max=None):

        self.ax.plot(np.log10(nu), np.log10(nuFnu))
        self.ax.set_xlabel(r'log($ \nu $)  (Hz)')
        self.ax.set_ylabel(r'log($ \nu F_{\nu} $ )  (erg cm$^{-2}$  s$^{-1}$)')
        self.ax.set_ylim(y_min, y_max)
        self.update_plot()



class  PlotSeedPhotons (BasePlot):

    def __init__(self):
        super(PlotSeedPhotons, self).__init__()

    def plot(self,nu,nuFnu,y_min=None,y_max=None):

        self.ax.plot(np.log10(nu), np.log10(nuFnu))
        self.ax.set_xlabel(r'log($ \nu $)  (Hz)')
        self.ax.set_ylabel(r'log(n )  (photons cm$^{-3}$  Hz$^{-1}$  ster$^{-1}$)')
        self.ax.set_ylim(y_min, y_max)
        self.update_plot()



import  pytest
from .base_class import TestBase


class TestModelFit(TestBase):

    def test(self,plot=False):
        self.integration_suite(plot=plot,sed_number=1)

    def integration_suite(self,plot=False,sed_number=1,phenom_dict=None,use_dep_pars=False,use_ebl=False,skip_minuit=False):
        if sed_number is not None:
            template,jet,sed_data=self.prepare_model(plot=plot,sed_number=sed_number)
        elif phenom_dict is not None:
            template = phenom_dict['template']
            jet = phenom_dict['jet']
            sed_data = phenom_dict['sed_data']
        else:
            raise RuntimeError('please provide either sed_number or  phenom_dict')

        fit_model=self.prepare_fit_model(jet=jet,use_dep_pars=use_dep_pars,use_ebl=use_ebl,template=template)
        fit_model, model_minimizer,sed_data=self.model_fit(fit_model=fit_model,
                                                            sed_data=sed_data, 
                                                            minimizer='lsb')


        if skip_minuit is False:
            fit_model, model_minimizer,sed_data=self.model_fit(fit_model=fit_model,
                                                                sed_data=sed_data, 
                                                                minimizer='minuit')

        return dict(fit_model=fit_model, model_minimizer=model_minimizer,sed_data=sed_data)

    def prepare_model(self, plot=True,sed_number=1):
        from .test_phenom_constr import TestPhenomenologyConstr
        t=TestPhenomenologyConstr()
        phenom_dict=t.integration_suite(sed_number=sed_number,plot=plot)
        prefit_jet=phenom_dict['jet']
        sed_data=phenom_dict['sed_data']
        template=phenom_dict['template']
        prefit_jet.set_gamma_grid_size(200)
        return template,prefit_jet,sed_data

    def prepare_fit_model(self,jet,use_dep_pars=False,use_ebl=False,template=None,minimizer='lsb'):
        from jetset.model_manager import FitModel

    
        fit_model = FitModel(jet=jet, name='SSC-best-fit-minuit', template=template)

        if use_ebl is True:
            from jetset.template_2Dmodel import EBLAbsorptionTemplate
            ebl_franceschini = EBLAbsorptionTemplate.from_name('Franceschini_2008')
            ebl_franceschini.show_model()
            fit_model.add_component(ebl_franceschini)
            fit_model.link_par(par_name='z_cosm', from_model='Franceschini_2008', to_model=jet.name)

        fit_model.freeze('jet_leptonic','z_cosm')
        fit_model.jet_leptonic.parameters.beam_obj.fit_range = [5, 50]
        fit_model.jet_leptonic.parameters.R_H.val = 5E17
        fit_model.jet_leptonic.parameters.R_H.frozen = False
        fit_model.jet_leptonic.parameters.R_H.fit_range = [1E15, 1E19]
        fit_model.jet_leptonic.parameters.R.fit_range = [10 ** 15.5, 10 ** 17.5]
        
        if use_dep_pars is True:
            if fit_model.parameters.get_par_by_name('jet_leptonic','B0') is None:
                fit_model.jet_leptonic.add_user_par(name='B0', units='G', val=1E3, val_min=0, val_max=None)
            
            if fit_model.parameters.get_par_by_name('jet_leptonic','R0') is None:
                fit_model.jet_leptonic.add_user_par(name='R0', units='cm', val=5E13, val_min=0, val_max=None)
            
            fit_model.jet_leptonic.parameters.R0.frozen = True
            fit_model.jet_leptonic.parameters.B0.frozen = True

            if fit_model.jet_leptonic.parameters.B._is_dependent  is False:
                fit_model.jet_leptonic.make_dependent_par(par='B', depends_on=['B0', 'R0', 'R_H'], par_expr='B0*(R0/R_H)')


        if use_ebl is True:
            if template is not None:
                fit_model.composite_expr = '(%s+host_galaxy)*Franceschini_2008'%jet.name
            else:
                fit_model.composite_expr = '(%s)*Franceschini_2008' % jet.name
        else:
            if template is not None:
                fit_model.composite_expr = '(%s+host_galaxy)'%jet.name
          

       

        if template is not None:
            fit_model.host_galaxy.parameters.nuFnu_p_host.frozen = False
            fit_model.host_galaxy.parameters.nu_scale.frozen = True

        
        return fit_model
        

    def model_fit(self,fit_model,sed_data, plot=False,minimizer='lsb'):
        
        from jetset.minimizer import ModelMinimizer
        
        model_minimizer = ModelMinimizer(minimizer)
        
        if minimizer == 'minuit':
            fit_model.jet_leptonic.parameters.gmin.fit_range = [2, 300]
            fit_model.jet_leptonic.parameters.gmax.fit_range = [1E5, 1E7]

        best_fit = model_minimizer.fit(fit_model, sed_data, 10 ** 11., 10 ** 29.0,
                                                    fitname='SSC-best-fit-%s'%minimizer, repeat=1)
        best_fit.show_report()
        best_fit.save_report('best-fit-%s-report.pkl'%minimizer)
        best_fit.bestfit_table.write('best-fit-%s-report.ecsv'%minimizer,overwrite=True)
        model_minimizer.save_model('model_minimizer_%s.pkl'%minimizer)

        model_minimizer = ModelMinimizer.load_model('model_minimizer_%s.pkl'%minimizer)
        fit_model.save_model('fit_model_%s.pkl'%minimizer)

        return fit_model, model_minimizer,sed_data

   


import  pytest
from .base_class import TestBase
import numpy as np

class TestTempEv(TestBase):

    def integration_suite(self,plot=False):
        self._all(plot=plot)

    def test_emp_ev_two_zone_cooling_acc(self, plot=False):
        print('--------> test_emp_ev_two_zone_cooling_acc',plot)
        from jetset.jet_emitters_factory import InjEmittersFactory
        from jetset.jet_model import Jet
        jet_model=Jet()
        q_inj=InjEmittersFactory().create_inj_emitters('pl',emitters_type='electrons',normalize=True)
        q_inj.parameters.gmin.val=9
        q_inj.parameters.gmax.val=10
        q_inj.parameters.p.val=0.5

        jet_model.parameters.beam_obj.val=30
        jet_model.parameters.B.val=0.2
        jet_model.parameters.z_cosm.val=0.03
        jet_model.parameters.R.val=5E15

        flare_duration=1.0E5
        duration=flare_duration*10
        t_D0=1.5E5
        t_A0=2.5E4
        T_esc_rad=1E60
        L_inj=5.0E39
        E_acc_max=4E60
        Delta_R_acc_ratio=0.1
        B_ratio=1.0
        T_SIZE=2E4
        NUM_SET=500
        Diff_Index=2.0
        Acc_Index=1.0

        from jetset.jet_timedep import JetTimeEvol
        temp_ev_acc=JetTimeEvol(jet_rad=jet_model,Q_inj=q_inj,inplace=True)

        temp_ev_acc.rad_region.jet.nu_min=1E8
        temp_ev_acc.acc_region.jet.nu_min=1E8
        T_SIZE=np.int(T_SIZE)

        if Delta_R_acc_ratio is not None:
            temp_ev_acc.parameters.Delta_R_acc.val=temp_ev_acc.parameters.R_rad_start.val*Delta_R_acc_ratio

        T_esc_acc=t_A0/(temp_ev_acc.parameters.Delta_R_acc.val/3E10)*2



        temp_ev_acc.parameters.duration.val=duration
        temp_ev_acc.parameters.TStart_Acc.val=0
        temp_ev_acc.parameters.TStop_Acc.val=flare_duration
        temp_ev_acc.parameters.TStart_Inj.val=0
        temp_ev_acc.parameters.TStop_Inj.val=flare_duration
        temp_ev_acc.parameters.T_esc_acc.val=T_esc_acc
        temp_ev_acc.parameters.T_esc_rad.val=T_esc_rad
        temp_ev_acc.parameters.t_D0.val=t_D0
        temp_ev_acc.parameters.t_A0.val=t_A0
        temp_ev_acc.parameters.Esc_Index_acc.val=Diff_Index-2
        temp_ev_acc.parameters.Esc_Index_rad.val=0
        temp_ev_acc.parameters.Acc_Index.val=Acc_Index
        temp_ev_acc.parameters.Diff_Index.val=Diff_Index
        temp_ev_acc.parameters.t_size.val=T_SIZE
        temp_ev_acc.parameters.num_samples.val=NUM_SET
        temp_ev_acc.parameters.E_acc_max.val=E_acc_max
        temp_ev_acc.parameters.L_inj.val=L_inj


        temp_ev_acc.parameters.gmin_grid.val=1.0
        temp_ev_acc.parameters.gmax_grid.val=1E8
        temp_ev_acc.parameters.gamma_grid_size.val=1500

        temp_ev_acc.parameters.B_acc.val=temp_ev_acc.rad_region.jet.parameters.B.val*B_ratio
        temp_ev_acc.init_TempEv()


        only_injection=True
        do_injection=True
        plot_fit_model=True
        plot_fit_distr=True
        plot_emitters=True
        plot_lcs=True
        delta_t_out=1000
        eval_cross_time=False
        rest_frame='obs'
        temp_ev_acc.run(only_injection=only_injection,
                        do_injection=do_injection,
                        cache_SEDs_acc=True, 
                        cache_SEDs_rad=True)

        if plot is True:
            p=temp_ev_acc.plot_tempev_emitters(region='rad',loglog=False,energy_unit='gamma',pow=0)
            p.ax.axvline(temp_ev_acc.temp_ev.gamma_eq_t_A, ls='--')
            p.ax.axvline(temp_ev_acc.temp_ev.gamma_eq_t_DA, ls='--')
            p.setlim(x_max=1E7,x_min=1,y_min=1E-18,y_max=100)

            p=temp_ev_acc.plot_tempev_emitters(region='acc',loglog=False,energy_unit='gamma',pow=0)
            p.ax.axvline(temp_ev_acc.temp_ev.gamma_eq_t_A, ls='--')
            p.ax.axvline(temp_ev_acc.temp_ev.gamma_eq_t_DA, ls='--')
            p.setlim(x_max=1E7,x_min=1,y_min=1E-30,y_max=100)

            p=temp_ev_acc.plot_tempev_model(region='rad',sed_data=None, use_cached = True)
            p.setlim(y_min=1E-18,x_min=1E7)


        lg=temp_ev_acc.rad_region.make_lc(nu1=2.4E22,nu2=7.2E25,name='gamma',eval_cross_time=False,delta_t_out=100,use_cached=True,frame='obs')

    
        if plot is True:
            import matplotlib.pyplot as plt
            plt.plot(lg['time'],lg['flux'])
            plt.xlabel('time (%s)'%lg['time'].unit)
            plt.ylabel('flux (%s)'%lg['flux'].unit)
        
        temp_ev_acc.save_model('two_zone_rad_acc.pkl')
        temp_ev_acc_1=JetTimeEvol.load_model('two_zone_rad_acc.pkl')

        if plot is True:
            p=temp_ev_acc_1.plot_tempev_model(region='rad',sed_data=None, use_cached = True)
        
        lx=temp_ev_acc_1.rad_region.make_lc(nu1=1E17,nu2=1E18,name='X',eval_cross_time=False,delta_t_out=100,use_cached=True,frame='obs')

        if plot is True:
            plt.plot(lx['time'],lx['flux'])
            plt.xlabel('time (%s)'%lg['time'].unit)
            plt.ylabel('flux (%s)'%lg['flux'].unit)

        print('-------->',plot)


__author__ = "Andrea Tramacere"



from .data_loader import log_to_lin, lin_to_log

from scipy.interpolate import interp1d


import numpy as np

import os

from .spectral_shapes import SED
from .plot_sedfit import PlotSED,PlotSpecComp

from .model_parameters import ModelParameter,ModelParameterArray
from .base_model import  Model

__all__=['SpectralTemplateLogLog', 'TemplateParameter', 'BigBlueBumpTemplateLogLog', 'HostGalaxyTemplateLogLog']

class TemplateParameter(ModelParameter):
    """
    This class is a subclass of the :class:`.ModelParameter` class,
    extending the base class to Template parameters.
    """
    def __init__(self,template,**keywords):
        
        self.template=template

        self.allowed_par_types=['nu-scale','nuFnu-scale']


        super(TemplateParameter,self).__init__(  **keywords)

        if 'val' in keywords.keys():
            val=keywords['val']
            self.assign_val(self.name,val)
            
        
        

    def set(self,**keywords):
        super(TemplateParameter,self).set(**keywords )
        
        """
        sets a parameter value checking for physical boundaries 
        """

        if 'val' in keywords.keys():
            self.assign_val(self.name,keywords['val']) 
        
    
    def assign_val(self,name,val):

        setattr(self.template,name,val)
        


class SpectralTemplateLogLog(Model):
    """
    Class to handle spectral templates
    """
    def __init__(self,template_type,cosmo,z=None,nu_size=100,name='TemplateModel'):
        """
        """
        super(SpectralTemplateLogLog, self).__init__(name=name)


        
        if z is not None  :
            self.z=z
            self.z_scale=np.log10(1+z)
        else:
            self.z=0.0
            self.z_scale=0.0

        self._scale = 'log-log'

        self.allowed_templates= self.get_allowed_template_name()
        
        self.nu_size=nu_size
        
        self.name = template_type
        
        self.model_type='template'
            
        self.SED = SED(name=self.name)
    
        self.parameters = ModelParameterArray()
      
        self.cosmo=cosmo

        self.DL=self.cosmo.get_DL_cm(self.z)
        self.flux_plot_lim=1E-30

    @staticmethod
    def get_allowed_template_name():
        return ['BBB','host_galaxy','user_defined']

    @classmethod
    def template_factory(cls,template_type,cosmo,z=None,nu_size=100,name='TemplateModel'):
        if template_type == 'BBB':
            return BigBlueBumpTemplateLogLog(template_type, cosmo, z=z, nu_size=nu_size, name=name)


        elif template_type == 'host_galaxy':

            return HostGalaxyTemplateLogLog(template_type, cosmo, z=z, nu_size=nu_size, name=name)

        else:
            raise ValueError("Wrong template type=%s, allowed=" % (template_type, cls.get_allowed_template_name()))

    def plot_model(self,plot_obj=None,clean=False,label=None,sed_data=None,color=None, density=False,frame='obs'):
        plot_obj = self._set_up_plot(plot_obj, sed_data, frame, density)

        if clean is True:
            plot_obj.clean_model_lines()


        if label is None:
            label=self.name

        plot_obj.add_model_plot(self.SED, line_style='-', label=label, flim=self.flux_plot_lim,color=color, density=density, frame=frame)

        return plot_obj


    def get_redshift(self):
        return self.z

    def set_Lum(self,nuFnu_p):
        self.L_D=self.get_L_D(nuFnu_p)

    
    def get_Lum(self,nuFnu_p):
        return 4*np.pi*self.DL*self.DL*nuFnu_p
        


    def load_teplate_from_file(self, file_path):
        """
        Loads a template from a file
        """
        xy = np.loadtxt(file_path, usecols=(0, 1), unpack=False, dtype=float)
        xy = xy[np.argsort(xy[:, 0])]
        self.nu_template = xy[:, 0]
        self.nuFnu_template = xy[:, 1]
        
   
    def log_func(self,nu_log):
        
        x_shift=getattr(self,self.x_scale)
        y_shift=getattr(self,self.y_scale)
        if np.shape(nu_log)==():
                nu=np.array([nu_log])
        
        
        x_log=nu_log-x_shift

        model=np.ones(x_log.size)+np.log10(self.flux_plot_lim)
            
        msk = x_log >self.nu_template.min() 
        msk*= x_log <self.nu_template.max()
            
        
        model[msk] = self.interp_func(x_log[msk])+y_shift
        
        return model
    
    
    def eval(self,fill_SED=True,nu=None,get_model=False,loglog=False):
        """
        Evaluates the Template for the current parameters values
        """    

        if nu is None:

            nu = np.copy(self.nu_template)


        if loglog is False:
            log_nu=np.log10(nu)

            lin_nu=nu
        else:
            log_nu=nu

            lin_nu=np.power(10.,log_nu)
        
        
        log_model= self.log_func(log_nu)
        
        model=np.power(10.,log_model)

        if fill_SED is True:
            self._fill(lin_nu = lin_nu, lin_model = model)

        if get_model is  True:
            if loglog is False:
            
                return model
            else:
                
                return log_model

        else:
            return None
            
  
class BigBlueBumpTemplateLogLog(SpectralTemplateLogLog):


    def __init__(self,template_type,cosmo,z=None,nu_size=100,name='TemplateModel'):

        super(BigBlueBumpTemplateLogLog, self).__init__(template_type,
                                                     cosmo,
                                                     z=z,
                                                     nu_size=nu_size,
                                                     name=name)

        self.Templates_dir = os.path.dirname(__file__) + '/Spectral_Templates_Repo'



        file_path = '%s/QSO_BBB_Template.dat' % self.Templates_dir

        self.load_teplate_from_file(file_path)
        self.nu_p_template = 10 ** self.nu_template[self.nuFnu_template.argmax()]
        # set nu_template according to redshift
        self.nu_template = self.nu_template - self.z_scale

        # add parameter
        self.nuFnu_p_BBB = 1.0
        self.nu_scale = 0.0
        self.parameters.add_par(
            TemplateParameter(self, name='nuFnu_p_BBB', par_type='nuFnu-scale', val=0.0, val_min=-20.0, val_max=20.0,
                              units='erg cm^-2 s^-1'))
        self.parameters.add_par(
            TemplateParameter(self, name='nu_scale', par_type='nu-scale', val=0.0, val_min=-2.0, val_max=2.0,
                              units='Hz'))
        self.y_scale = 'nuFnu_p_BBB'
        self.x_scale = 'nu_scale'
        self.nuLnu_p_BBB = self.get_Lum(self.nuFnu_p_BBB)
        self.T_BBB = self.get_T_BBB()
        self.interp_func = interp1d(self.nu_template, self.nuFnu_template)

    def get_T_BBB(self):
        return self.nu_p_template / (1.39 * 5.879e10)

    def set_BBB_pars(self, fit_model, model_name):
        self.DL = self.cosmo.get_DL_cm(self.z)

        nuFnu_p, nuFnu_p_err = log_to_lin(
            log_val=fit_model.parameters.get(model_name, 'nuFnu_p_BBB', 'best_fit_val'),
            log_err=fit_model.parameters.get(model_name, 'nuFnu_p_BBB', 'best_fit_err'))
        err_rel = nuFnu_p_err / nuFnu_p
        self.nuLnu_p_BBB = self.get_Lum(nuFnu_p)
        self.nuLnu_p_BBB_err = self.nuLnu_p_BBB * err_rel

        self.nu_p, self.nu_p_err = log_to_lin(
            log_val=fit_model.parameters.get(model_name, 'nu_scale', 'best_fit_val'),
            log_err=fit_model.parameters.get(model_name, 'nu_scale', 'best_fit_err'))
        err_rel = self.nu_p_err / self.nu_p
        self.nu_p = self.nu_p_template * self.nu_p
        self.nu_p_err = self.nu_p * err_rel

        self.T_Disk = self.get_T_BBB()
        self.T_Disk_err = self.T_Disk * err_rel


class HostGalaxyTemplateLogLog(SpectralTemplateLogLog):

    def __init__(self, template_type, cosmo, z=None, nu_size=100, name='TemplateModel'):

        super(HostGalaxyTemplateLogLog, self).__init__(template_type,
                                                     cosmo,
                                                     z=z,
                                                     nu_size=nu_size,
                                                     name=name)



        self.Templates_dir = os.path.dirname(__file__) + '/Spectral_Templates_Repo'

        file_path = '%s/HostgalaxyTemplate.dat' % self.Templates_dir

        self.load_teplate_from_file(file_path)
        self.nu_p_template = 10 ** self.nu_template[self.nuFnu_template.argmax()]
        # set nu_template according to redshift
        self.nu_template = self.nu_template - self.z_scale

        # add parameter
        self.nuFnu_p_host = 1.0
        self.nu_scale = 0.0

        self.parameters.add_par(
            TemplateParameter(self, name='nuFnu_p_host', par_type='nuFnu-scale', val=0.0, val_min=-20.0, val_max=20.0,
                              units='erg cm^-2 s^-1'))
        self.parameters.add_par(
            TemplateParameter(self, name='nu_scale', par_type='nu-scale', val=0.0, val_min=-2.0, val_max=2.0,
                              units='Hz'))

        self.y_scale = 'nuFnu_p_host'
        self.x_scale = 'nu_scale'
        self.nuLnu_p_host = self.get_Lum(self.nuFnu_p_host)

        self.interp_func = interp1d(self.nu_template, self.nuFnu_template)



    def set_host_pars(self, fit_model, model_name):
        self.DL = self.cosmo.get_DL_cm(self.z)

        nuFnu_p, nuFnu_p_err = log_to_lin(
            log_val=fit_model.parameters.get(model_name, 'nuFnu_p_host', 'best_fit_val'),
            log_err=fit_model.parameters.get(model_name, 'nuFnu_p_host', 'best_fit_err'))
        err_rel = nuFnu_p_err / nuFnu_p

        self.nuLnu_p_host = self.get_Lum(nuFnu_p)
        self.nuLnu_p_host_err = self.nuLnu_p_host * err_rel

        err_rel = self.nuLnu_p_host_err / self.nuLnu_p_host
        self.nuLnu_p_host = 4 * np.pi * self.DL * self.DL * self.nuFnu_p_host
        self.nuLnu_p_host_err = self.nuLnu_p_host * err_rel

        self.nu_p, self.nu_p_err = log_to_lin(
            log_val=fit_model.parameters.get(model_name, 'nu_scale', 'best_fit_val'),
            log_err=fit_model.parameters.get(model_name, 'nu_scale', 'best_fit_err'))
        err_rel = self.nu_p_err / self.nu_p
        self.nu_p = self.nu_p * self.nu_p
        self.nu_p_err = self.nu_p * err_rel

    def load_teplate_from_file(self, file_path):
        """
        Loads a template from a file
        """
        xy = np.loadtxt(file_path, usecols=(0, 1), unpack=False, dtype=float)
        xy = xy[np.argsort(xy[:, 0])]
        self.nu_template = xy[:, 0]
        self.nuFnu_template = xy[:, 1]


import pytest
import numpy as np
def test(plot=False):
    print('--------> test_hadronic_jet',plot)
    from jetset.jet_model import Jet
    j = Jet(proton_distribution='plc')
    j.parameters.gmin.val = 2
    j.parameters.gmax.val = 1E8
    j.parameters.NH_pp.val = 1E10
    j.parameters.N.val = 1E1
    j.parameters.B.val = 80

    j.parameters.p.val = 2.5
    j.eval()
    j.show_model()

    j.eval()
    print('--------> j.energetic_report')
    j.energetic_report(verbose=False)
    assert('U_p_cold' not in j.energetic_dict.keys())
    np.testing.assert_allclose(j.energetic_dict['U_p'],j.emitters_distribution.eval_U(),rtol=1E-2)

    sum1=j.spectral_components.Sum.SED.nuFnu
    if plot is True:
        j.plot_model()
    j.save_model('test_jet_hadronic.pkl')
    j_new = Jet.load_model('test_jet_hadronic.pkl')
    j_new.eval()
    sum2 = j_new.spectral_components.Sum.SED.nuFnu
    np.testing.assert_allclose(sum2,sum1, rtol=1E-5)
    print('-------->  j_new.energetic_report')
    j_new.energetic_report(verbose=False)
    assert('U_p_cold' not in j.energetic_dict.keys())   
    np.testing.assert_allclose(j.energetic_dict['U_p'],j.emitters_distribution.eval_U(),rtol=1E-2)



__author__ = "Andrea Tramacere"

from scipy.optimize import leastsq
import scipy as s
from numpy import polyfit,polyval,polyder
import  numpy as np

__all__=['check_maxima','cubic_peaks','do_cubic_fit','do_cubic_fit_peaks','do_linear_fit',
         'do_log_Parab_FIT','filter_interval','find_max_cubic','lin','parab','residuals_cubic_peaks',
         'residuals_linfit','residuals_parab']


def filter_interval(x,x_range):
    msk1=x>=x_range[0]
    msk2=x<=x_range[1]
    return msk1*msk2




#-------- LINEAR FIT----------------------------
def do_linear_fit(x,y,dy=None,x_range=None):
    
    if x_range is not None  :
        msk = filter_interval(x,x_range)
        x_fit=x[msk]
        y_fit=y[msk]
        dy_fit=dy[msk] 
    else:
        x_fit=x
        y_fit=y 
        dy_fit=dy
    
    if dy is None:
        dy_fit=s.ones(len(x_fit))
   
        


    if len(x_fit)<2:
        raise RuntimeError("fit failed number of points less then 2")
     
    #print nu_fit,nuFnu_fit,err_nuFnu_fit
    pinit = [1.0,10.0]
    #print "input", log_nu,log_nuLnu    
    p,cov,infodict,mesg,ier = leastsq(residuals_linfit, pinit,args=(x_fit,y_fit,dy_fit), full_output=1)
    
    chisq=(infodict['fvec']**2).sum()
    dof=len(x_fit)-len(p)
    if dof>0:
        rmse=s.sqrt(chisq/dof)
    else:
        rmse=0
        
    if cov is not None  :
        err=s.sqrt(cov.diagonal())*rmse
    else:
        err=s.zeros(len(p))
    #print p
    return p,err

def lin(p,x):

    return p[0]+p[1]*(x)

def residuals_linfit(p,x,y,dy):
    #print len(dy),len(x),len(y),len(p)
    err = (y-lin(p,x))/dy
    return err
#-----------------------------------------------


#----CUBIC-PEAKS------------------------------------
def cubic_peaks(p,x):
    #p0=yp
    #p1=b
    #p2=xp
    #
    return p[0] + p[1]*(x-p[2])*(x-p[2]) + p[3]*(x-p[2])*(x-p[2])*(x-p[2])


def residuals_cubic_peaks(p,x,y,dy,template):
    if template is not None  :
        model1=s.power(10,cubic_peaks(p[:4],x))
        model2=s.power(10,template.func(x,nuFnu_scale=p[4]))
        model=s.log10(model1+model2)
    else:
        model=cubic_peaks(p[:4],x)
    
    
    
    err = (y-model)/dy
    return err

def do_cubic_fit_peaks(x,y,xp,yp,dy=None,x_range=None,template=None):
    
    if x_range is not None  :
        msk = filter_interval(x,x_range)
        x_fit=x[msk]
        y_fit=y[msk]
        dy_fit=dy[msk] 
    else:
        x_fit=x
        y_fit=y 
        dy_fit=dy
    
    if dy is None:
        dy_fit=s.ones(len(x_fit))
    
    
    if template is not None  :
        pinit = [yp,0,xp,0,yp]
    else:
        pinit = [yp,0,xp,0]
    
    
    p,cov,infodict,mesg,ier = leastsq(residuals_cubic_peaks, pinit,args=(x_fit,y_fit,dy_fit,template), full_output=1)
    
    #print"cubic fit", out[0]
    chisq=(infodict['fvec']**2).sum()
    # dof is degrees of freedom
    dof=len(x_fit)-len(p)
    if dof>0:
        rmse=s.sqrt(chisq/dof)
    else:
        rmse=0
        
    #print "--> chi_red", rmse
    if cov is not None  :
        err=s.sqrt(cov.diagonal())*rmse
    else:
        err=s.zeros(len(p))
    #print p[3]
    #print "--> fit log: ",mesg,ier
    #print infodict['fvec']
    par_parab=[p[2],p[0],p[1]]
    err_parab=[err[2],err[0],err[1]]
    if len(p)==5:
        par_parab.append(p[4])
        err_parab.append(err[4])
        
    return par_parab,err_parab,dof
#------------------------------------------------------


#-----CUBIC   -----------------------------------------
def find_max_cubic(p,x,y,dy=None,x_range=None):
    der=polyder(p,1)
    delta=der[1]*der[1] -(4*der[0]*der[2])
    if delta>=0:
        #print p,der
        root1=(-der[1]+s.sqrt(delta))/(2*der[0])
        root2=(-der[1]-s.sqrt(delta))/(2*der[0])
        #print root1,root2
        if der[0]>0:
            xp=min(root1,root2)
        else:
            xp=max(root1,root2)
        
        yp=polyval(p,xp)
        #p0=yp
        #p1=xp
        #p2=b
        #print "xp,yp=",xp,yp
        p_peaks,err,dof=do_cubic_fit_peaks(x,y,xp,yp,dy=dy,x_range=x_range)
        return  p_peaks,err,dof
    else:
        print ("!!! no maxima found for cubic fit")
        xp=None
        yp=None
        b=None
        return [yp,xp,b],[0,0,0],'No'


def check_maxima(xm,ym,b,x,y):
    c = np.array([x, y]).T
    c = c[np.argsort(c[:, 1])]
    max=c[-1]
    #print "check max", c
    #print max[0],max[1]
    #min=c[0]


    if b<0:
       return max[0],max[1]
    if b>=0:
        if ym<max[1]:
            return max[0],max[1]
        else:
            return xm,ym


def do_cubic_fit(x,y,dy=None,x_range=None):
    
    if x_range is not None  :
        msk = filter_interval(x,x_range)
        x_fit=x[msk]
        y_fit=y[msk]
        dy_fit=dy[msk] 
    else:
        x_fit=x
        y_fit=y 
        dy_fit=dy
    
    if dy is None:
        dy_fit=s.ones(len(x_fit))
    
    
    
    
    p=polyfit(x_fit,y_fit,3)
        
    val,err,dof=find_max_cubic(p,x_fit,y_fit,dy=dy_fit)
    return p,val,err
#-----------------------------------------------

#---LOG PAR-----------------------------------------

def do_log_Parab_FIT(x,y,xp,yp,beta,dy=None,x_range=None,template=None):

    if x_range is not None  :
        msk = filter_interval(x,x_range)
        x_fit=x[msk]
        y_fit=y[msk]
        dy_fit=dy[msk] 
    else:
        x_fit=x
        y_fit=y 
        dy_fit=dy
    
    if dy is None:
        dy_fit=s.ones(len(x_fit))
    
    
  
    
    if template is not None  :
        pinit = [xp,yp,beta,yp]
    else:
        pinit = [xp,yp,beta]
 
    p,cov,infodict,mesg,ier = leastsq(residuals_parab, pinit,args=(x_fit, y_fit,dy_fit,template), full_output=1)
     
    chisq=(infodict['fvec']**2).sum()

    dof=len(x_fit)-len(p)
    
    if dof>0:
        rmse=s.sqrt(chisq/dof)
    else:
        rmse=0
        
    
    if cov is not None  :
        err=s.sqrt(cov.diagonal())*rmse
    else:
        err=s.zeros(len(p))
    
    
    #print p,err
    return p,err

def parab(p,x):
    
    return p[1]+p[2]*(x-p[0])*(x-p[0])


def residuals_parab(p,x,y,dy,template):
    #print p
    if template is not None  :
        model1=s.power(10,parab(p[:3],x))
        model2=s.power(10,template.interp_template(template.func,x,p[3]))
        model=s.log10(model1+model2)
    else:
        model=parab(p,x)

    err = (y-model)/dy
    return err

#---------------------------------------------------


__author__ = "Andrea Tramacere"

import os
try:
    from sherpa.models.model import ArithmeticModel, modelCacher1d, RegriddableModel1D
    from sherpa.models.parameter import Parameter
    from sherpa import data
    from sherpa.fit import Fit
    from sherpa.stats import Chi2
    from sherpa.optmethods import LevMar
    from sherpa.fit import Fit
    from sherpa.stats import Chi2
    from sherpa import data as sherpa_data

    sherpa_installed = True

except:
    #on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
    
    #if on_rtd is True:
    #    pass
    #else:
    raise  ImportError('to use sherpa plugin you need to install sherpa: https://sherpa.readthedocs.io/en/latest/install.html')

import  numpy as np
from .plot_sedfit import  PlotSED
from .minimizer import  Minimizer

__all__=['JetsetSherpaModel','plot_sherpa_model']


class JetsetSherpaModel(RegriddableModel1D):
    """
    authomatic sherpa model generator
    """

    def __init__(self, jetset_model,par_list=None,clone=False):
        if clone is True:
            self._jetset_model = jetset_model.clone()
        else:
            self._jetset_model=jetset_model

        self._jp_list = []
        self._jp_par_array = []
        self._jp_list_names = []
        setattr(self, '_jetset_ncalls', 0)
        keep=True
        #print('-->, ', par_list)
        for p in self._jetset_model.parameters.par_array:
            if par_list is None:
                keep = True
            else:
                keep = p in par_list
            #print('-->, ',p, p.name, keep)
            if p is not None and keep is True:

                if p.name.lower() in self._jp_list_names or p.name.upper() in self._jp_list_names:
                    name = p.name + '_sh'
                    print('jetset model name', p.name, 'renamed to ', name, 'due to sherpa internal naming convention')
                else:
                    name = p.name
                if p.fit_range_min is not None:
                    val_min=p.fit_range_min
                else:
                    val_min = p.val_min

                if p.fit_range_max is not None:
                    val_max = p.fit_range_max
                else:
                    val_max = p.val_max

                sh_p = Parameter(self._jetset_model.name, name, p.val, min=val_min, max=val_max, units=p.units)
                setattr(self, sh_p.name, sh_p)
                p._sherpa_ref = sh_p
                if np.isnan(sh_p.max):
                    sh_p.max = sh_p.hard_max
                if np.isnan(sh_p.min):
                    sh_p.min = sh_p.hard_min

                self._jp_list.append(sh_p)
                self._jp_par_array.append(p)
                self._jp_list_names.append(p.name)
        RegriddableModel1D.__init__(self, jetset_model.name,(p._sherpa_ref for p in self._jp_par_array))


    def calc(self, pars, x):
        for ID, p in enumerate(self._jp_list):
            j_p = self._jp_par_array[ID]
            j_p.val = p.val
        self._jetset_ncalls +=1
        return self._jetset_model.eval(get_model=True, nu=x)

    def plot_model(self, fit_range, model_range=[1E10, 1E30], nu_grid_size=200, plot_obj=None, sed_data=None):
        self._jetset_model.set_nu_grid(model_range[0], model_range[1], nu_grid_size)
        self._jetset_model.eval()
        plot_obj = self._jetset_model.plot_model(plot_obj=plot_obj, sed_data=sed_data)
        plot_obj.add_model_residual_plot(data=sed_data, model=self._jetset_model,
                                         fit_range=[fit_range[0], fit_range[1]])


def plot_sherpa_model(sherpa_model, fit_range=None, model_range=[1E10, 1E30], nu_grid_size=200, sed_data=None,
                      add_res=False, plot_obj=None, label=None, line_style=None):
    if fit_range is not None:
        x = np.logspace(np.log10(fit_range[0]), np.log10(fit_range[1]), nu_grid_size)
    else:
        x = np.logspace(np.log10(model_range[0]), np.log10(model_range[1]), nu_grid_size)
    y = sherpa_model(x)

    if plot_obj is None:
        plot_obj = PlotSED(frame='obs', density=False)

    if sed_data is not None:
        plot_obj.add_data_plot(sed_data=sed_data)

    plot_obj.add_xy_plot(x, y, label=label, line_style=line_style)

    if add_res is True and fit_range is not None:
        nufnu_res = sherpa_model(sed_data.data['nu_data'])
        y_res = (sed_data.data['nuFnu_data'] - nufnu_res) / sed_data.data['dnuFnu_data']
        x_res = sed_data.data['nu_data']
        plot_obj.add_xy_residual_plot(x=x_res, y=y_res, fit_range= [fit_range[0], fit_range[1]])

    return plot_obj


class SherpaMinimizer(Minimizer):

    def __init__(self, model,method=LevMar(),stat=Chi2()):
        if sherpa_installed is True:
            pass
        else:
            raise ImportError('sherpa not installed, \n to use sherpa plugin you need to install sherpa: https://sherpa.readthedocs.io/en/latest/install.html')

        super(SherpaMinimizer, self).__init__(model)
        self._method=method
        self._stat=stat
        self._sherpa_model = None
        self._sherpa_data = None
        self.pbar = None

    def _create_sherpa_model(self):
        self._sherpa_model = JetsetSherpaModel(jetset_model = self.model.fit_model, par_list=self.model.fit_par_free)

    def _create_sherpa_data(self):
        self._sherpa_data = sherpa_data.Data1D("sed", self.model.data['x'], self.model.data['y'], staterror=self.model.data['dy'])

    @property
    def sherpa_fitter(self):
        return self._sherpa_fitter

    @property
    def calls(self):
        if self._sherpa_model is not None:
            return self._sherpa_model._jetset_ncalls
        else:
            return None

    @calls.setter
    def calls(self,n):
        if self._sherpa_model is not None:
            self._sherpa_model._jetset_ncalls = n


    def _fit(self, max_ev,):
        self._create_sherpa_model()
        self._create_sherpa_data()
        self._sherpa_model._jetset_ncalls = 0


        self._sherpa_fitter=Fit(self._sherpa_data,self._sherpa_model, method=self._method,stat=self._stat)

        self.mesg  = self._sherpa_fitter.fit()
        self.covar = self.mesg.covar
        self.pout = [p for p in self.mesg.parvals]
        self.p = [p for p in self.mesg.parvals]

    def _set_fit_errors(self):
        self.errors = [np.sqrt(np.fabs(self.covar[pi, pi])) for pi in range(len(self.model.fit_par_free))]


__author__ = "Andrea Tramacere"

import re
import inspect
import sys
import warnings
import os
import json



__all__=['check_frame','unexpected_behaviour']



def check_frame(frame):
    allowed=['obs','src','blob']
    if frame not in allowed:
        raise RuntimeError('rest frame', frame, 'not allowed',allowed)

def unexpected_behaviour():
    raise RuntimeError('the code reached a condition that should never happen!')


def clean_var_name(s):
    _s = s
    s.replace('-', '_')
    s.replace(' ', '_')

    # Remove invalid characters
    s = re.sub('[^0-9a-zA-Z_]', '', s)

    # Remove leading characters until we find a letter or underscore
    s = re.sub('^[^a-zA-Z_]+', '', s)

    if len(s) == 0:
        raise RuntimeError('the string ', _s, 'is not valid, please change it (do not use leading numbers or math symbols)')
    return s


class NoTraceBackWithLineNumber(Exception):
    def __init__(self, msg):
        try:
            ln = sys.exc_info()[-1].tb_lineno
        except AttributeError:
            ln = inspect.currentframe().f_back.f_lineno
        self.args = "{0.__name__} (line {1}): {2}".format(type(self), ln, msg),
        sys.exit(self)

def new_version_warning():
    m = '\n\n' + '*'*80 + '\n'
    m+= 'Something wrong has happened. Please, look at the exception message.\n'
    m+= '*' * 80 + '\n'
    warnings.warn(m)


def parameters_warning():
    pass


def old_model_warning():
    m = '\n\n' + '*'*80 + '\n'
    m+= 'you are loading a model supported for version<1.1.0, starting from version 1.1.0 \n'
    m+= 'the saved model has changed,  plase update to the new model the new format, \n'
    m += 'by saving it with this version\n'
    m+= '*' * 80 + '\n'
    warnings.warn(m)

class JetkerneltException(Exception):

    def __init__(self, message='Jeset  exception', debug_message=''):
        super(JetkerneltException, self).__init__(message)
        self.message=message
        self.debug_message=debug_message


#def str_hook(pairs):
#    new_pairs = []
#    for key, value in pairs:
#        if isinstance(value, unicode):
#            value = value.encode('utf-8')
#        if isinstance(key, unicode):
#            key = key.encode('utf-8')
#        new_pairs.append((key, value))
#    return dict(new_pairs)



def safe_run(func):

    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
           message =  'the jetkernel failed\n'
           message += '\n exception message: \n'
           message += '%s'%str(e)
           new_version_warning()

           raise JetkerneltException(message=message)


    return func_wrapper


def set_str_attr(obj,name,val):
    #print('set obj', obj,'name',name ,'to', val)
    try:

        try:
            setattr(obj, name,val)
        except:
            setattr(obj, name, val.encode('ascii'))
    except Exception as e:
        raise RuntimeError('error setting attr',name,'execption:',e)



def get_info():
    with open(os.path.dirname(__file__) + '/pkg_info.json') as fp:
        _info = json.load(fp)

    return _info



import  pytest
from .base_class import TestBase

class TestPhenomenologyConstr(TestBase):

    def integration_suite(self,sed_number=1,plot=False):
        from .test_data import TestData
        t=TestData()
        sed_data=t.test_data_loader(sed_number=sed_number,plot=False)
        my_shape=self.test_spectral_indices(sed_data=sed_data,plot=plot,sed_number=sed_number)
        my_shape=self.test_sed_shaper(my_shape=my_shape,sed_data=sed_data,sed_number=sed_number)
        prefit_jet, my_shape,template=self.test_model_constr(my_shape=my_shape,sed_data=sed_data,plot=plot,sed_number=sed_number)
        
        return dict(jet=prefit_jet, shape=my_shape,template=template,sed_data=sed_data)

    def test_spectral_indices(self,sed_data=None,plot=True,sed_number=1):
        from jetset.sed_shaper import SEDShape
        from .test_data import TestData
        if sed_data is None:
            t=TestData()
            sed_data=t.test_data_loader(sed_number=sed_number,plot=False)

        my_shape = SEDShape(sed_data)
        my_shape.eval_indices(silent=True)
        if plot is True:
            p = my_shape.plot_indices()
            p.setlim(y_min=1E-15, y_max=1E-6)

        return my_shape

    def test_sed_shaper(self,my_shape=None, plot=True, sed_number=1,sed_data=None,check_host_gal_template=False):
        if my_shape is None:
            my_shape=self.test_spectral_indices(plot=plot,sed_number=sed_number,sed_data=sed_data)
        
        if sed_number == 2:
            check_host_gal_template=True
        
        print('==> check_host_gal_template',check_host_gal_template,sed_number)
        mm, best_fit = my_shape.sync_fit(check_host_gal_template=check_host_gal_template,
                                        Ep_start=None,
                                        minimizer='lsb',
                                        silent=True,
                                        fit_range=[10, 21])

        best_fit.show_report()
        best_fit.save_report('synch_shape_fit_rep.pkl')
        best_fit.bestfit_table.write('IC_shape_fit_rep.ecsv',overwrite=True)
        mm, best_fit= my_shape.IC_fit(fit_range=[23, 29], minimizer='minuit',silent=True)

        if plot is True:
            p = my_shape.plot_shape_fit()
            p.setlim(y_min=1E-15)

        best_fit.show_report()
        best_fit.save_report('IC_shape_fit_rep.pkl')
        best_fit.bestfit_table.write('IC_shape_fit_rep.ecsv',overwrite=True)
        my_shape.save_values('sed_shape_values.ecsv')
        return my_shape

    def test_model_constr(self,my_shape=None, plot=True,sed_number=1,sed_data=None,check_host_gal_template=False):
        from jetset.obs_constrain import ObsConstrain

        if sed_number == 2:
            check_host_gal_template=True
        print('==> check_host_gal_template', check_host_gal_template, sed_number)
        if my_shape is None:
            my_shape=self.test_sed_shaper(plot=plot,sed_number=sed_number,sed_data=sed_data,check_host_gal_template=check_host_gal_template)

        sed_obspar = ObsConstrain(beaming=25,
                                B_range=[0.001, 0.1],
                                distr_e='lppl',
                                t_var_sec=3 * 86400,
                                nu_cut_IR=1E11,
                                SEDShape=my_shape)

        prefit_jet = sed_obspar.constrain_SSC_model(electron_distribution_log_values=False,silent=True)
        prefit_jet.save_model('prefit_jet.pkl')
        template=None
        if sed_number == 2:
            template = my_shape.host_gal
        return prefit_jet, my_shape, template




__author__ = "Andrea Tramacere"


#Peak values
Sync_nuFnu_p_dic={'obs':'nuFnu_peak_Sync_obs'}
Sync_nuFnu_p_dic['src']='nuLnu_peak_Sync_src'
Sync_nuFnu_p_dic['blob']='nuLnu_peak_Sync_blob'


Sync_nu_p_dic={'obs':'nu_peak_Sync_obs'}
Sync_nu_p_dic['src']='nu_peak_Sync_src'
Sync_nu_p_dic['blob']='nu_peak_Sync_blob'


SSC_nuFnu_p_dic={'obs':'nuFnu_peak_SSC_obs'}
SSC_nuFnu_p_dic['src']='nuLnu_peak_SSC_src'
SSC_nuFnu_p_dic['blob']='nuLnu_peak_SSC_blob'


SSC_nu_p_dic={'obs':'nu_peak_SSC_obs'}
SSC_nu_p_dic['src']='nu_peak_SSC_src'
SSC_nu_p_dic['blob']='nu_peak_SSC_blob'

# nu_src_start_stop_dict={'Sync':['nu_start_Sync', 'nu_start_Sync']}
# nu_src_start_stop_dict['SSC']=['nu_start_SSC', 'nu_stop_SSC']
# nu_src_start_stop_dict['EC_BLR']=['', '']
# nu_src_start_stop_dict['EC_DT']=['', '']
# nu_src_start_stop_dict['EC_Disk']=['', '']
# nu_src_start_stop_dict['EC_CMB']=['', '']
# nu_src_start_stop_dict['EC_CMB_stat']=['', '']
# nu_src_start_stop_dict['Bremss_ep']=['', '']
# nu_src_start_stop_dict['PP_gamma']=['', '']
# nu_src_start_stop_dict['PP_neutrino_tot']=['', '']
# nu_src_start_stop_dict['PP_neutrino_e']=['', '']
# nu_src_start_stop_dict['PP_neutrino_mu']=['', '']






#Spectral components
nuFnu_obs_dict={'Sum':['nuFnu_sum_grid', 'nu_grid']}
nuFnu_obs_dict['Sync']=['nuFnu_Sync_grid', 'nu_grid']
nuFnu_obs_dict['SSC']=['nuFnu_SSC_grid', 'nu_grid']
nuFnu_obs_dict['EC_BLR']=['nuFnu_EC_BLR_grid', 'nu_grid']
nuFnu_obs_dict['EC_DT']=['nuFnu_EC_DT_grid', 'nu_grid']
nuFnu_obs_dict['DT']=['nuFnu_DT_grid', 'nu_grid']
nuFnu_obs_dict['Star']=['nuFnu_Star_grid', 'nu_grid']
nuFnu_obs_dict['EC_Disk']=['nuFnu_EC_Disk_grid', 'nu_grid']
nuFnu_obs_dict['Disk']=['nuFnu_Disk_grid', 'nu_grid']
nuFnu_obs_dict['EC_CMB']=['nuFnu_EC_CMB_grid', 'nu_grid']
nuFnu_obs_dict['EC_CMB_stat']=['nuFnu_EC_CMB_stat_grid', 'nu_grid']
nuFnu_obs_dict['Bremss_ep']=['nuFnu_bremss_ep_grid', 'nu_grid']
nuFnu_obs_dict['PP_gamma']=['nuFnu_pp_gamma_grid', 'nu_grid']
nuFnu_obs_dict['PP_neutrino_tot']=['nuFnu_pp_neutrino_tot_grid', 'nu_grid']
nuFnu_obs_dict['PP_neutrino_e']=['nuFnu_pp_neutrino_e_grid', 'nu_grid']
nuFnu_obs_dict['PP_neutrino_mu']=['nuFnu_pp_neutrino_mu_grid', 'nu_grid']






#seed-photon fields components
n_seed_dic={'DT':['n_DT','nu_DT']}
n_seed_dic['EC_DT']=['n_DT','nu_DT']

n_seed_dic['EC_Disk']=['n_Disk','nu_Disk']
n_seed_dic['Disk']=['n_Disk','nu_Disk']
n_seed_dic['EC_BLR']=['n_BLR','nu_BLR']
n_seed_dic['EC_CMB']=['n_CMB','nu_CMB']
n_seed_dic['EC_S']=['n_CMB','nu_CMB']
n_seed_dic['SSC']=['n_Sync','nu_Sync']
n_seed_dic['Star']=['n_Star','nu_Sync']


#nuLnu_dic={'SUM':['','nu_']}
#nuLnu_dic['Sync']=['nuFnu_Sync','nu_']
#nuLnu_dic['SSC']=['nuFnu_Sync','nu_']
#nuLnu_dic['EC_BLR']=['nuFnu_Sync','nu_']
#nuLnu_dic['EC_DT']=['nuFnu_Sync','nu_']
#nuLnu_dic['EC_Disk']=['nuFnu_Sync','nu_']





#Electron distributions

gamma_dic_e={'electron_distr':['Ne','griglia_gamma_Ne_log']}

gamma_dic_p={'proton_distr':['Np','griglia_gamma_Np_log']}
gamma_dic_pp_e_second={'e_second_inj':['Q_inj_e_second','griglia_gamma_Ne_log']}

s_dic={'pl':'p'}
s_dic['lppl']='s'
s_dic['lp']='s'
s_dic['bkn']='p'
s_dic['plc']='p'

s1_dic={'bkn':'p_1'}

gamma_cut_dic={'pl':'gmax'}
gamma_cut_dic['lppl']='gamma0_log_parab'
gamma_cut_dic['lp']='gmax'
gamma_cut_dic['plc']='gamma_cut'

r_dic={'pl':'r'}
r_dic['lppl']='r'
r_dic['lp']='r'
r_dic['lpep']='r'

gamma_3p_dic={'lpep':'gammap_log_parab'}
gamma_3p_dic['bkn']='gamma_break'


available_N_distr=['lp', 'pl', 'lppl', 'lpep', 'plc', 'bkn', 'spitkov', 'lppl_pile_up', 'bkn_pile_up']
available_N_distr_descr=['log-parabola',
                         'powerlaw',
                         'log-parabola with low-energy powerlaw branch',
                         'log-parabola defined by peak energy',
                         'powerlaw with cut-off',
                         'broken powerlaw',
                         'spitkov',
                         'log-parabola with low-energy powerlaw branch and pile-up',
                         'broken powerlaw and pileup']


available_emitters_type=['electrons','protons']

N_distr_descr={}
for m,d in zip(available_N_distr,available_N_distr_descr):
    N_distr_descr[m]=d


allowed_disk_type=['BB','MultiBB','Mono']



__author__ = "Andrea Tramacere"

import os
import glob

__all__=[]


_kernel_dir = os.path.join(os.path.dirname(__file__), 'mathkernel')
bessel_table_file_path = os.path.join(_kernel_dir, 'F_Sync.dat')




__author__ = "Andrea Tramacere"

import numpy as np
from astropy import units
from .frame_converter import  convert_nuFnu_to_nuLnu_src, convert_nu_to_src, convert_nu_src_to_nu_blob, convert_nuLnu_src_to_nuLnu_blob
from .utils import *

__all__=['SED']





class SED(object):
    """
    Class handling the SED 
    """
    def __init__(self,
                 name=None,
                 nu=None,
                 nuFnu=None,
                 nu_residuals=None,
                 residuals=None,
                 nu_src_residuals=None,
                 nuLnu_src_residuals=None,
                 dl=None,
                 z=None,
                 log_log=False,
                 beaming=None):

        if beaming is None:
            beaming =1

        self.beaming=beaming

        self.name=name

        self._nu_units=units.Hz
        self._nuFnu_units=units.erg/units.cm**2/units.s

        self._nu_src_units=units.Hz
        self._nuLnu_src_units = units.erg/units.s
        self._nuLnu_blob_units  = units.erg/units.s


        self.nu=(nu)

        self.nuFnu=(nuFnu)

        self._loglog=log_log

        if z is not None and dl is not None:
            # calling setter do not change
            self.nu_src= (z)

            # calling setter do not change
            self.nuLnu_src = (z,dl)

        self.nu_residuals=nu_residuals
        self.residuals=residuals

        self.nu_src_residuals = nu_src_residuals
        self.nuLnu_src_residuals = nuLnu_src_residuals

    @property
    def nu(self):
        return self._nu

    @nu.setter
    def nu(self,nu):
        if nu is None:
            self._nu=nu
        else:
            self._nu=nu*self._nu_units


    @property
    def nuFnu(self):
        return self._nuFnu


    @nuFnu.setter
    def nuFnu(self,nuFnu):
        if nuFnu is None:
            self._nuFnu=nuFnu
        else:
            self._nuFnu = nuFnu * self._nuFnu_units

    @property
    def nu_src(self):
        return self._nu_src

    @nu_src.setter
    def nu_src(self, z):
        #print('->self._nu',self._nu)
        if self._nu is None:
            self._nu_src = self._nu
        else:
            if self._loglog is True:
                self._nu_src =np.log10( convert_nu_to_src(10**(self._nu.value),z,in_frame='obs') )* self._nu_units
            else:
                self._nu_src = convert_nu_to_src(self._nu.value,z,in_frame='obs') * self._nu_units

        #print('->self._nu_src', self._nu_src)

    @property
    def nuLnu_src(self):
        return self._nuLnu

    @nuLnu_src.setter
    def nuLnu_src(self, t):
        z,dl=t
        #print('2')
        #print('->',t,z,dl,self._loglog)
        if self._nuFnu is None:
            self._nuLnu = None
        else:
            if self._loglog is True:
                self._nuLnu =np.log10( convert_nuFnu_to_nuLnu_src(10**(self._nuFnu.value),z,'obs',dl)) * self._nuLnu_src_units
            else:
                self._nuLnu = convert_nuFnu_to_nuLnu_src(self._nuFnu.value, z, 'obs', dl) * self._nuLnu_src_units

    @property
    def nuLnu_blob(self):
        return convert_nuLnu_src_to_nuLnu_blob(self._nuLnu, beaming=self.beaming, in_frame='src')* self._nuLnu_src_units

    @property
    def nu_blob(self):
        return convert_nu_src_to_nu_blob(self._nu_src, beaming=self.beaming, in_frame='src')* self._nu_units

    def get_model_points(self,log_log=False,frame='obs'):

        check_frame(frame)
        if frame == 'obs':
            x, y = self.nu.value, self.nuFnu.value
        elif frame == 'src':
            x, y = self.nu_src.value, self.nuLnu_src.value
        elif frame == 'blob':
            x,y = self.nu_blob.value, self.nuLnu_blob.value
        else:
            unexpected_behaviour()

        if log_log is True:
            x=np.copy(x)
            y=np.copy(y)
            msk = y>0
            x = np.log10(x)
            y[~msk] = -1E10
            y[msk]= np.log10(y[msk])

        return x, y

    def get_residuals(self, log_log=False):

        residuals = self.residuals
        nu_residuals = self.nu_residuals

        if log_log == False:
            return nu_residuals, residuals
        else:
            return nu_residuals, np.log10(residuals)

    def fill(self,nu=None,nuFnu=None,nu_residuals=None,residuals=None,log_log=False):

        self._loglog=log_log
        #if nu is not None:
        self.nu=(nu)
        
        #if nuFnu is not None:
        self.nuFnu=(nuFnu)
         
        if residuals is not None:
            
            self.residuals=residuals
         
        if nu_residuals is not None:
            
            self.nu_residuals=nu_residuals

    def fill_nuLnu(self, nu_src_residuals=None, nuLnu_src_residuals=None, z=None, dl=None):


        if z is not None and dl is not None:

            #calling setter do not change
            self.nu_src = (z)

            # calling setter do not change
            self.nuLnu_src = (z,dl)

        if nuLnu_src_residuals is not None:
            self.nuLnu_src_residuals = nuLnu_src_residuals

        if nu_src_residuals is not None:
            self.nu_src_residuals = nu_src_residuals
            
#
# class poly_shape(object):
#     """
#     Class for log-log polynomial shapes
#     """
#     def __init__(self,name=None,nu=None,nuFnu=None):
#         self.name=name
#         self.nu=nu
#         self.nuFnu=nuFnu
#
#     def get_model_points(self):
#         return self.nu,self.nuFnu
#
#





__author__ = "Andrea Tramacere"


import warnings

import numpy as np

from  . import minimizer



from .model_parameters import  CompositeModelParameterArray

from .spectral_shapes import  SED
   
from .base_model import  Model

from .plot_sedfit import  PlotSED

from .utils import  clean_var_name

from .jet_model import Jet

from .cosmo_tools import  Cosmo

import  dill as pickle

__all__=['FitModel']

class CompositeModelContainer(object):

    def __init__(self):
        self._components_list=[]
        self._components_value=[]
        self._components_value_dict = {}
        self.parameters=CompositeModelParameterArray()

    def add_component(self, model_comp,fit_model):
        try:
            assert (model_comp.name not in [m_comp.name for m_comp in self._components_list])
        except Exception as e:
            raise RuntimeError('model name:', model_comp.name, 'already assigned',e)

        try:
            assert (model_comp not in self._components_list)
        except Exception as e:
            raise RuntimeError('model:', model_comp, 'already added',e)

        self._components_list.append(model_comp)
        self._components_value.append(None)

        self._components_value_dict[model_comp.name] = self._components_value[-1]

        self.parameters.add_model_parameters(model_comp)
        setattr(self, clean_var_name(model_comp.name), model_comp)
        setattr(fit_model, clean_var_name(model_comp.name), model_comp)

    @property
    def components_list(self):
        return self._components_list

    def get_model_by_name(self, model_name,get_idx=False):
        model = None
        idx=None
        for ID,m in enumerate(self._components_list):
            if m.name == model_name:
                model = m
                idx=ID
        if model is None:
            warnings.warn('Model',model_name,'not present')
        if get_idx is False:
            return model
        else:
            return model,idx

    def del_component(self,model_name,fit_model):
        m,ID=self.get_model_by_name(model_name,get_idx=True)
        if m is not None:
            _p = self._components_list.pop(ID)
            _p = self._components_value.pop(ID)
            _p = self.parameters.del_model_parameters(m)
            del self._components_value_dict[m]

        delattr(self,model_name)
        delattr(fit_model, model_name)

    def show_pars(self):
        self.parameters.show_pars()

    def show_model(self):
        for c in self._components_list:
            #print()
            c.show_model()
            #print()





class FitModel(Model):
    """
    """
    
    def __init__(self,
                 elec_distr=None,
                 jet=None,
                 name='no-name',
                 out_dir=None,
                 flag=None,
                 template=None,
                 loglog_poly=None,
                 analytical=None,
                 nu_size=200,
                 cosmo=None,
                 composite_expr=None,
                 **keywords):

        super(FitModel,self).__init__( model_type='composite_model', **keywords)
       
        if  jet is not None and elec_distr is not None:
            #!! warning or error?
            raise RuntimeError("you can't provide both elec_distr and jet, only one")

        self.sed_data=None
        self.nu_min_fit=1E6
        self.nu_max_fit=1E30
        self.name=name
        self.SED=SED(name=self.name)
        self.nu_min=1E6
        self.nu_max=1E30
        self.nu_size=nu_size
        self.flux_plot_lim=1E-30
        self.components=CompositeModelContainer()
        self.parameters=self.components.parameters
        self.composite_expr = composite_expr

        if elec_distr is not None:
            jet=Jet(cosmo=cosmo,name=flag, electron_distribution=elec_distr, jet_workplace=None)
            self.add_component(jet)
        
        if jet is not None:
            self.add_component(jet)

        if cosmo is None:
          self.cosmo=Cosmo()
          warnings.warn('no cosmology defined, using default %s'%self.cosmo)
        else:
            self.cosmo = cosmo

        if template is not None:
            self.add_component(template)

        if loglog_poly is not None:
            self.add_component(loglog_poly)

        if analytical is not None:
            self.add_component(analytical)

    def plot_model(self,plot_obj=None,clean=False,sed_data=None,frame='obs',skip_components=False,label=None,skip_sub_components=False, density=False):
        plot_obj=self._set_up_plot(plot_obj,sed_data,frame,density)

        if clean is True:
            plot_obj.clean_model_lines()

        if skip_components is False:
            line_style = '--'

            for mc in self.components._components_list:
                comp_label = mc.name
                if hasattr(mc,'SED'):
                    try:
                        plot_obj.add_model_plot(mc.SED, line_style=line_style,label=comp_label,flim=self.flux_plot_lim, density=density, frame=frame)
                    except Exception as e:
                        try:
                            mc.eval()
                            plot_obj.add_model_plot(mc.SED, line_style=line_style, label=comp_label,
                                                    flim=self.flux_plot_lim, density=density, frame=frame)
                        except Exception as e:
                            raise RuntimeError('for model', mc.name, "problem with plotting SED", e)


                if skip_sub_components is False:
                    if hasattr(mc,'spectral_components_list'):
                        for c in mc.spectral_components_list:
                            comp_label = c.name
                            if comp_label!='Sum':
                                if hasattr(c, 'SED'):
                                    try:
                                        plot_obj.add_model_plot(c.SED, line_style=line_style, label='  -%s'%comp_label, flim=self.flux_plot_lim, density=density, frame=frame)
                                    except Exception as e:
                                        try:
                                            #print('==> reval',mc.name)
                                            mc.eval()
                                            _c=mc.spectral_components.get_spectral_component_by_name(c.name)
                                            plot_obj.add_model_plot(_c.SED, line_style=line_style, label='  -%s'%comp_label, flim=self.flux_plot_lim, density=density, frame=frame)
                                        except Exception as e:
                                            raise RuntimeError('for model', mc.name, "spectral component",c.name, "problem with plotting SED", e)
        line_style = '-'
        if label is None:
            label=self.name

        plot_obj.add_model_plot(self.SED, line_style=line_style, label=label, flim=self.flux_plot_lim,fit_range=[self.nu_min_fit,self.nu_max_fit], density=density, frame=frame  )
        plot_obj.add_model_residual_plot(data=sed_data, model=self, fit_range=[self.nu_min_fit, self.nu_max_fit])

        #if frame == 'src' and sed_data is not None:
        #    sed_data.z = z_sed_data

        return plot_obj



    def set_nu_grid(self,nu_min=None,nu_max=None,nu_size=None):
        if nu_size is not None:
            self.nu_size=nu_size
        
        if nu_min is not None:
            self.nu_min=nu_min
        
        if nu_max is not None:
            self.nu_max=nu_max
        
        for model_comp in self.components._components_list:
        
            if nu_size is not None:
                model_comp.nu_size=nu_size
            
            if nu_min is not None:
                model_comp.nu_min=nu_min
            
            if nu_max is not None:
                model_comp.nu_max=nu_max

    def set(self,model,par_name, *args, **kw):
        self.parameters.set(model, par_name, *args, **kw)

    def set_par(self,model,par_name,val):
        self.parameters.set(model, par_name, val=val)

    def get(self,model,par_name,*args):
        self.parameters.get(model,par_name,*args)

    def get_par_by_name(self,model,par_name):
        return self.parameters.get_par_by_name(model, par_name)

    def freeze(self,model,par_name):
        self.parameters.freeze(model,par_name)

    def free(self,model,par_name):
        self.parameters.free(model,par_name)

    def free_all(self,):
        self.parameters.free_all()

    def freeze_all(self,):
        self.parameters.freeze_all()

    def add_component(self,m):
        self.components.add_component(m,self)

    def del_component(self,m):
        self.components.del_component(m,self)

    @property
    def composite_expr(self):
        return self._composite_expr

    def link_par(self,par_name,from_model,to_model):
        if isinstance(from_model, list) is False:
            from_model = [from_model]
        self.parameters.link_par(par_name,from_model, to_model)

    @composite_expr.setter
    def composite_expr(self,expr_string):
        if expr_string is None:
            self._composite_expr = expr_string
        else:
            try:
                for key, val in self.components._components_value_dict.items():
                    exec(key + '= 1.0')
                eval(expr_string)
            except Exception as e:
                raise RuntimeError('function string not valid',e)

        self._composite_expr=expr_string

    def _eval_composite_func(self,loglog):
        #transform each key into a local var
        for key, val in self.components._components_value_dict.items():
            if loglog is True:
                exec(key + '=np.power(10.,**val)')
            else:
                exec(key + '=val')

        return eval(self.composite_expr)

    def _eval_model(self, lin_nu, log_nu, loglog,fill_SED):
        lin_model = np.zeros(lin_nu.size)
        log_model = None
        for model_comp in self.components._components_list:
            model_comp.cosmo=self.cosmo
            #print('--> eval composite component', model_comp.name)
            if loglog is False:
                self.components._components_value_dict[model_comp.name]=model_comp.eval(nu=lin_nu, fill_SED=fill_SED, get_model=True, loglog=loglog)
            else:
                self.components._components_value_dict[model_comp.name] = model_comp.eval(nu=log_nu, fill_SED=fill_SED,get_model=True, loglog=loglog)
            if self.composite_expr is None:
                if loglog is False:
                    lin_model += self.components._components_value_dict[model_comp.name]
                else:
                    lin_model += np.power(10.,self.components._components_value_dict[model_comp.name])

        if self.composite_expr is not None:
            lin_model = self._eval_composite_func(loglog)

        #lin_model[lin_model< self.flux_plot_lim]=self.flux_plot_lim
        if loglog is True:
            log_model=np.log10(lin_model)

        return lin_model, log_model

    def eval(self,nu=None,fill_SED=True,get_model=False,loglog=False,label=None,phys_output=False):

        out_model= None
        #print('--> model mananger eval 1')
        lin_nu, log_nu = self._prepare_nu_model(nu, loglog)
        #print('--> model mananger eval 2',lin_nu[0],log_nu[0] )
        lin_model,log_model  = self._eval_model(lin_nu, log_nu, loglog, fill_SED)

        if fill_SED is True:
            self._fill(lin_nu, lin_model)

        if get_model is True:
            if loglog is True:
                out_model = log_model
            else:
                out_model = lin_model

        return out_model

    @classmethod
    def load_model(cls, file_name):
         c = pickle.load(open(file_name, "rb"))
         return cls._build_model(c)
    
    @staticmethod
    def _build_model(c):
        try:
            ml=c.components.components_list[::]
            for m in ml:
                try:
                    c.del_component(m.name)
                except:
                    pass
            for m in ml:
                c.add_component(m)

            for p in c.parameters.par_array:
                if p._linked is True:
                    p._linked = False
                    p._is_dependent = False
                    #print(p.name,p._root_par,[p.model],p._linked_root_model,p.immutable)
                    c.parameters.link_par(p._root_par.name,[p.model.name],p._linked_root_model.name)

            for m in c.components.components_list:
                if isinstance(m,Jet):
                    m._fix_par_dep_on_load()
            if isinstance(c, Model):
                c.eval()
                return c
            else:
                raise RuntimeError('The model you loaded is not valid please check the file name')

        except Exception as e:
            raise RuntimeError(e)

    def clone(self):
        return self._build_model(pickle.loads(pickle.dumps(self, protocol=pickle.HIGHEST_PROTOCOL)))

    def show_model_components(self):
        print("")
        print('-'*80)

        print("Composite model description")
        print('-'*80)
        print("name: %s  " % (self.name))
        print("type: %s  " % (self.model_type))
        print("components models:")
        for m in self.components._components_list:
            print(' -model name:', m.name, 'model type:', m.model_type)
        print('')
        print('-'*80)

    def show_model(self):
        print("")
        print('-'*80)

        print("Composite model description")
        print('-'*80)
        print("name: %s  " % (self.name))
        print("type: %s  " % (self.model_type))
        print("components models:")
        for m in self.components._components_list:
            print(' -model name:',m.name,'model type:', m.model_type)
        print('')
        print('-'*80)

        print("individual component description")

        self.components.show_model()
        print('-'*80)




import  pytest
from .base_class import TestBase

class TestData(TestBase):

    def integration_suite(self,plot=False):
        self._all(plot=plot)

    def test_data_loader(self,plot=True,sed_number=2):
        from jetset.data_loader import ObsData, Data
        from jetset.test_data_helper import test_SEDs

        data = Data.from_file(test_SEDs[0])
        sed_data = ObsData(data_table=data)
        sed_data.show_data_sets()
        sed_data.filter_data_set('-1', exclude=True)
        sed_data.show_data_sets()
        sed_data.reset_data()
        sed_data.show_data_sets()
        sed_data.filter_data_set('-1', exclude=False)
        sed_data.show_data_sets()
        sed_data.reset_data()

        data=Data.from_file(test_SEDs[sed_number])
        sed_data=ObsData(data_table=data)
        sed_data.group_data(bin_width=0.2)
        print('sed file->',test_SEDs[sed_number])
        sed_data.add_systematics(0.1,[10.**6,10.**29])
        if plot is True:
            p=sed_data.plot_sed()

        return sed_data
    

    def test_data_from_numpy(self,plot=False):
        pass

    def test_data_from_table(self,plot=False):
        pass




__author__ = "Andrea Tramacere"

from .model_parameters import ModelParameterArray, ModelParameter
from .base_model import Model

from .spectral_shapes import SED

from numpy import log10,power,sqrt,shape,zeros

import numpy as np

from numpy import polyfit,polyval,polyder


__all__=['find_max_cubic','LogCubic','LogLinear','LogLogModel','LogParabolaEp','LogParabolaPL','PolyParameter']

class PolyParameter(ModelParameter):
    """
    This class is a subclass of the :class:`.ModelParameter` class,
    extending the base class to loglog polynomial parameters.
    """
    def __init__(self,polymodel,**keywords):
        
        self.polymodel=polymodel

        self.allowed_par_types=['curvature','peak freq','peak flux','third-degree','spectral-slope','flux-const','turn-over freq']
        

        super(PolyParameter,self).__init__(  **keywords)

        if 'val' in keywords.keys():
            val=keywords['val']
            self.assign_val(self.name,val)
            
        
        
    def set(self,**keywords):
        super(PolyParameter,self).set(**keywords )
        
        """
        sets a parameter value checking for physical boundaries 
        """
        if 'val' in keywords.keys():
            self.assign_val(self.name,keywords['val']) 
        
    
    def assign_val(self,name,val):

        setattr(self.polymodel,name,val)



class LogLogModel(Model):
    
    def __init__(self,nu_size=100, **keywords):
       
        super(LogLogModel,self).__init__(**keywords)
        self.model_type='LogLogModel'
    
    
    def lin_func(self,nu):
        nu_log=np.log10(nu)
        return np.power(10.0,self.log_func(nu_log))
         
       
       

class LogLinear(LogLogModel):
    """
    Class to handle log-linear model
    """
    
    def __init__(self,nu_size=100,**keywords):
        """
        """
     
        super(LogLinear,self).__init__(  **keywords)
        
        self.name='LogLinear'
            
        self.SED = SED(name=self.model_type)
    
        self.parameters = ModelParameterArray()
      
        self.alpha=-1.0
        self.parameters.add_par(PolyParameter(self,name='alpha',par_type='spectral-slope',val=-1.0,val_min=-10.,val_max=10.,units=''))
        
       
        self.K=-10
        self.parameters.add_par(PolyParameter(self,name='K',par_type='flux-const',val=-10.0,val_min=None,val_max=None,units='erg cm^-2 s^-1',log=True))
        
         
        
    def log_func(self,log_nu):
    
            #x_log=log_nu
            #print(x_log)
            #print self.Ep,self.Sp,self.b,self.c,x_log
            
            return self.K + self.alpha*(log_nu)
    
    
    
    



class LogParabolaEp(LogLogModel):
    
    """
    Class to handle log-parabolic model
    """
    
    def __init__(self,nu_size=100, **keywords):
        """
        """
        
        super(LogParabolaEp,self).__init__(  **keywords)
        
        self.name='LogParabolaEp'
            
        self.SED = SED(name=self.model_type)
    
        self.parameters = ModelParameterArray()
      
        self.b=-1.0
        self.parameters.add_par(PolyParameter(self,name='b',par_type='curvature',val=-1.0,val_min=-10.,val_max=10.,units=''))
        
        self.Ep=14.0
        self.parameters.add_par(PolyParameter(self,name='Ep',par_type='peak freq',val=14.0,val_min=0.,val_max=30.,units='Hz',log=True))
        
        self.Sp=-10
        self.parameters.add_par(PolyParameter(self,name='Sp',par_type='peak flux',val=-10.0,val_min=-30.,val_max=0.,units='erg cm^-2 s^-1',log=True))
        
    def log_func(self,log_nu):
    
            x_log=log_nu-self.Ep

            
            return self.Sp + self.b*(x_log*x_log)
    
    
   




class LogParabolaPL(LogLogModel):
    """
    Class to handle a log-par + pl model
    """
    
    def __init__(self,nu_size=100,**keywords):
        """
        """
        
        super(LogParabolaPL,self).__init__(  **keywords)
        
        self.name='LogParabolaPL'
            
        self.SED = SED(name=self.model_type)
    
        self.parameters = ModelParameterArray()
      
        self.b=-1.0
        self.parameters.add_par(PolyParameter(self,name='b',par_type='curvature',val=-1.0,val_min=-10.,val_max=10.,units=''))

        self.E0=14.0
        self.parameters.add_par(PolyParameter(self,name='E0',par_type='turn-over freq',val=14.0,val_min=0.,val_max=30.,units='Hz',log=True))
        
        self.alpha=0.5
        self.parameters.add_par(PolyParameter(self,name='alpha',par_type='spectral-slope',val=0.5,val_min=-10.,val_max=10.,units=''))

        self.K=-10
        self.parameters.add_par(PolyParameter(self,name='K',par_type='flux-const',val=-10.0,val_min=-30.,val_max=0.,units='erg cm^-2 s^-1',log=True))
        
    
    def log_func(self,log_nu):
        
        
        #TODO fix this to numpy array function
        if shape(log_nu)==():
            return self.composite_func(log_nu)

        else:
            y_log=zeros(log_nu.size)
            for i in range(log_nu.size):
                y_log[i]=self.composite_func(log_nu[i])

            return y_log
   
    def composite_func(self,log_nu):     
        
        x_log=log_nu-self.E0
        
        if log_nu >=  self.E0:
            
            return self.K + x_log*(self.alpha + self.b*x_log)

        else:    

            return  self.K + x_log*self.alpha 



    


class LogCubic(LogLogModel):
    
    """
    Class to handle log-cubic model
    """
    
    def __init__(self,nu_size=100, **keywords):
        """
        """
        
        
        super(LogCubic,self).__init__(  **keywords)
        
        
        
        self.name='LogCubic'
            
        self.SED = SED(name=self.model_type)
    
        self.parameters = ModelParameterArray()
      
        self.b=-1.0
        self.parameters.add_par(PolyParameter(self,name='b',par_type='curvature',val=-1.0,val_min=-10.,val_max=10.,units=''))
        
        self.c=0.0
        self.parameters.add_par(PolyParameter(self,name='c',par_type='third-degree',val=-1.0,val_min=-10.,val_max=10.,units=''))

        self.Ep=14.0
        self.parameters.add_par(PolyParameter(self,name='Ep',par_type='peak freq',val=14.0,val_min=0.,val_max=30.,units='Hz',log=True))
        
        self.Sp=-10
        self.parameters.add_par(PolyParameter(self,name='Sp',par_type='peak flux',val=-10.0,val_min=-30.,val_max=0.,units='erg cm^-2 s^-1',log=True))
        
        

    def log_func(self,log_nu):
    
    
            x_log=log_nu-self.Ep
            
            #print self.Ep,self.Sp,self.b,self.c,x_log
            
            return self.Sp + self.b*(x_log*x_log)+ self.c*(x_log*x_log*x_log)
            
    
    

def find_max_cubic(x_log,y_log,x_range=None):
    
    if x_range is not None:
        msk = x_log >x_range[0]
        msk*= x_log <x_range[1]
        
        x_log=x_log[msk]
        y_log=y_log[msk]
    
    p=polyfit(x_log,y_log,3)
    
    der=polyder(p,1)
    
    delta=der[1]*der[1] -(4*der[0]*der[2])
    if delta>=0:
        #print p,der
        root1=(-der[1]+sqrt(delta))/(2*der[0])
        root2=(-der[1]-sqrt(delta))/(2*der[0])
        #print root1,root2
        if der[0]>0:
            xp=min(root1,root2)
        else:
            xp=max(root1,root2)
        

        return  xp
    else:
        print("!!! no maxima found for cubic fit")

        return None


__author__ = "Andrea Tramacere"

import numpy as np
import numba as nb
import pprint
from jetset.jet_emitters import EmittersDistribution, InjEmittersDistribution
__all__=['EmittersFactory']

_available_dict = {'lp': 'log-parabola',
                   'pl': 'powerlaw',
                   'lppl': 'log-parabola with low-energy powerlaw branch',
                   'lpep': 'log-parabola defined by peak energy',
                   'plc': 'powerlaw with cut-off',
                   'bkn': 'broken powerlaw',
                   'superexp': 'powerlaw with super-exp cut-off'}


@nb.njit(fastmath=True, cache=True)
def distr_func_bkn(gamma_break, gamma, p, p_1):
    f = np.zeros(gamma.shape)
    m = gamma < gamma_break
    f[m] = np.power(gamma[m], -p)
    f[~m] = np.power(gamma_break, -(p - p_1)) * np.power(gamma[~m], -p_1)
    return f


@nb.njit(fastmath=True, cache=True)
def distr_func_pl(gamma, p, ):
    return np.power(gamma, -p)


@nb.njit(fastmath=True, cache=True)
def distr_func_plc(gamma, gamma_cut, p,):
    return np.power(gamma, -p) * np.exp(-(gamma / gamma_cut) )


@nb.njit(fastmath=True, cache=True)
def distr_func_super_exp(gamma, gamma_cut, p, a):
    return np.power(gamma, -p) * np.exp(-(1 / a) * (gamma / gamma_cut) ** a)


@nb.njit(fastmath=True, cache=True)
def distr_func_lp(gamma, gamma0_log_parab, r, s):
    return np.power((gamma / gamma0_log_parab), (-s - r * np.log10(gamma / gamma0_log_parab)))


@nb.njit(fastmath=True, cache=True)
def distr_func_lep(gamma, gamma_p, r):
    return np.power(10., (-r * np.power(np.log10(gamma / gamma_p), 2)))


@nb.njit(fastmath=True, cache=True)
def distr_func_lppl(gamma, gamma0_log_parab, r, s):
    f = np.zeros(gamma.shape)
    m = gamma < gamma0_log_parab
    f[m] = np.power(gamma[m]/gamma0_log_parab, -s)
    f[~m] = np.power(gamma[~m]/gamma0_log_parab, (-s -r*np.log10(gamma[~m] / gamma0_log_parab )))
    return f


class EmittersFactory:
    def __repr__(self):
        return str(pprint.pprint(self._available_dict))

    def __init__(self):
        self._available_dict=_available_dict

        self._func_dict = {'pl': self._create_pl,
                           'bkn': self._create_bkn,
                           'plc': self._create_plc,
                           'lp': self._create_lp,
                           'lppl': self._create_lppl,
                           'lpep': self._create_lpep,
                           'superexp': self._create_super_exp}

        self._set_emitters_class()

    def _set_emitters_class(self):
        if type(self) == EmittersFactory:
            self._emitters_class=EmittersDistribution
        elif type(self) ==InjEmittersFactory:
            self._emitters_class = InjEmittersDistribution
        else:
            raise RuntimeError('class instance', type(self), 'not valid')

    @staticmethod
    def available_distributions():
        for k in _available_dict.keys():
            print('%s: %s' % (k, _available_dict[k]))

    @staticmethod
    def available_distributions_list():
        return  _available_dict.keys()

    def create_emitters(self,
                        name,
                        gamma_grid_size=200,
                        log_values=False,
                        emitters_type='electrons',
                        normalize=True,
                        skip_build=False):

        if name not in self._available_dict.keys():
            raise RuntimeError('name', name, 'not among available', self._available_dict.keys())

        return self._func_dict[name](gamma_grid_size=gamma_grid_size,
                                     log_values=log_values,
                                     normalize=normalize,
                                     skip_build=skip_build,
                                     emitters_type=emitters_type)

    def _create_bkn(self,gamma_grid_size,log_values,normalize,skip_build,emitters_type):

        n_e_bkn = self._emitters_class(name='bkn',
                                       spectral_type='bkn',
                                       normalize=normalize,
                                       emitters_type=emitters_type,
                                       log_values=log_values,
                                       skip_build=skip_build,
                                       gamma_grid_size=gamma_grid_size)

        a_t, b_t = n_e_bkn.set_bounds(1, 1E9, log_val=n_e_bkn._log_values)
        gamma_break_val = n_e_bkn._set_log_val(1E4,log_val=log_values)
        n_e_bkn.add_par('gamma_break', par_type='turn-over-energy', val=gamma_break_val, vmin=a_t, vmax=b_t, unit='lorentz-factor',log=log_values)
        n_e_bkn.add_par('p', par_type='LE_spectral_slope', val=2.5, vmin=-10., vmax=10, unit='')
        n_e_bkn.add_par('p_1', par_type='HE_spectral_slope', val=3.5, vmin=-10., vmax=10, unit='')
        n_e_bkn.set_distr_func(distr_func_bkn)
        return n_e_bkn

    def _create_pl(self,gamma_grid_size, log_values, normalize, skip_build, emitters_type):

        n_e_pl = self._emitters_class(name='pl',
                                       spectral_type='pl',
                                       normalize=normalize,
                                       emitters_type=emitters_type,
                                       log_values=log_values,
                                       skip_build=skip_build,
                                       gamma_grid_size=gamma_grid_size)

        a_t, b_t = n_e_pl.set_bounds(1, 1E9, log_val=n_e_pl._log_values)

        n_e_pl.add_par('p', par_type='LE_spectral_slope', val=2.0, vmin=-10., vmax=10, unit='')
        n_e_pl.set_distr_func(distr_func_pl)

        return n_e_pl


    def _create_plc(self, gamma_grid_size,log_values,normalize,skip_build,emitters_type):

        n_e_plc = n_e_bkn = self._emitters_class(name='plc',
                                             spectral_type='plc',
                                             normalize=normalize,
                                             emitters_type=emitters_type,
                                             log_values=log_values,
                                             skip_build=skip_build,
                                             gamma_grid_size=gamma_grid_size)

        a_t, b_t = n_e_plc.set_bounds(1, 1E9, log_val=n_e_plc._log_values)
        gamma_cut_val = n_e_plc._set_log_val(1E4,log_val=log_values)
        n_e_plc.add_par('gamma_cut', par_type='turn-over-energy', val=gamma_cut_val, vmin=a_t, vmax=b_t,
                              unit='lorentz-factor',log=log_values)
        n_e_plc.add_par('p', par_type='LE_spectral_slope', val=2.0, vmin=-10., vmax=10, unit='')
        n_e_plc.set_distr_func(distr_func_plc)
        return n_e_plc


    def _create_super_exp(self, gamma_grid_size, log_values, normalize, skip_build, emitters_type):

        n_e_super_exp  = self._emitters_class(name='super_exp',
                                             spectral_type='plc',
                                             normalize=normalize,
                                             emitters_type=emitters_type,
                                             log_values=log_values,
                                             skip_build=skip_build,
                                             gamma_grid_size=gamma_grid_size)

        a_t, b_t = n_e_super_exp.set_bounds(1, 1E9, log_val=n_e_super_exp._log_values)
        gamma_cut_val = n_e_super_exp._set_log_val(1E4,log_val=log_values)
        n_e_super_exp.add_par('gamma_cut', par_type='turn-over-energy', val=gamma_cut_val, vmin=a_t, vmax=b_t,
                              unit='lorentz-factor',log=log_values)
        n_e_super_exp.add_par('p', par_type='LE_spectral_slope', val=2.0, vmin=-10., vmax=10, unit='')
        n_e_super_exp.add_par('a', par_type='spectral_curvature', val=1.0, vmin=0., vmax=100., unit='')
        n_e_super_exp.set_distr_func(distr_func_super_exp)

        return n_e_super_exp

    def _create_lp(self, gamma_grid_size,log_values,normalize,skip_build,emitters_type):

        n_lp  = self._emitters_class(name='lp',
                                    spectral_type='lp',
                                    normalize=normalize,
                                    emitters_type=emitters_type,
                                    log_values=log_values,
                                    skip_build=skip_build,
                                    gamma_grid_size=gamma_grid_size)

        a_t, b_t = n_lp.set_bounds(1, 1E9, log_val=n_lp._log_values)
        gamma0_log_parab_val = n_lp._set_log_val(1E4,log_val=log_values)
        n_lp.add_par('gamma0_log_parab', par_type='turn-over-energy', val=gamma0_log_parab_val, vmin=a_t, vmax=b_t,
                              unit='lorentz-factor',log=log_values)
        n_lp.add_par('s', par_type='LE_spectral_slope', val=2.0, vmin=-10., vmax=10, unit='')
        n_lp.add_par('r', par_type='spectral_curvature', val=0.4, vmin=-15., vmax=15., unit='')

        n_lp.set_distr_func(distr_func_lp)

        return n_lp

    def _create_lpep(self,gamma_grid_size, log_values, normalize, skip_build, emitters_type):

        n_lep  = self._emitters_class(name='lpep',
                                    spectral_type='lp',
                                    normalize=normalize,
                                    emitters_type=emitters_type,
                                    log_values=log_values,
                                    skip_build=skip_build,
                                    gamma_grid_size=gamma_grid_size)

        a_t, b_t = n_lep.set_bounds(1, 1E9, log_val=n_lep._log_values)
        gamma_p_val = n_lep._set_log_val(1E4,log_val=log_values)
        n_lep.add_par('gamma_p', par_type='turn-over-energy', val=gamma_p_val, vmin=a_t, vmax=b_t,
                     unit='lorentz-factor',log=log_values)
        n_lep.add_par('r', par_type='spectral_curvature', val=0.4, vmin=-15., vmax=15., unit='')
        n_lep.set_distr_func(distr_func_lep)

        return n_lep

    def _create_lppl(self, gamma_grid_size, log_values, normalize, skip_build, emitters_type):

        n_lppl  = self._emitters_class(name='lppl',
                                    spectral_type='lp',
                                    normalize=normalize,
                                    emitters_type=emitters_type,
                                    log_values=log_values,
                                    skip_build=skip_build,
                                    gamma_grid_size=gamma_grid_size)

        a_t, b_t = n_lppl.set_bounds(1, 1E9, log_val=n_lppl._log_values)
        gamma0_log_parab_val = n_lppl._set_log_val(1E4,log_val=log_values)

        n_lppl.add_par('gamma0_log_parab', par_type='turn-over-energy', val=gamma0_log_parab_val, vmin=a_t, vmax=b_t,
                     unit='lorentz-factor',log=log_values)
        n_lppl.add_par('s', par_type='LE_spectral_slope', val=2.0, vmin=-10., vmax=10, unit='')
        n_lppl.add_par('r', par_type='spectral_curvature', val=0.4, vmin=-15., vmax=15., unit='')

        n_lppl.set_distr_func(distr_func_lppl)

        return n_lppl


class InjEmittersFactory(EmittersFactory):

    def __init__(self):
        super(InjEmittersFactory, self).__init__()



    def create_inj_emitters(self,
                        name,
                        gamma_grid_size=200,
                        log_values=False,
                        emitters_type='electrons',
                        normalize=True,
                        skip_build=False):

        if name not in self._available_dict.keys():
            raise RuntimeError('name', name, 'not among available', self._available_dict.keys())
        return self._func_dict[name](gamma_grid_size=gamma_grid_size,
                                     log_values=log_values,
                                     normalize=normalize,
                                     skip_build=skip_build,
                                     emitters_type=emitters_type)